# -*- coding: utf-8 -*-
# This file is auto-generated, don't edit it. Thanks.
from typing import Dict
from Tea.core import TeaCore

from alibabacloud_tea_openapi.client import Client as OpenApiClient
from alibabacloud_tea_openapi import models as open_api_models
from alibabacloud_tea_util.client import Client as UtilClient
from alibabacloud_endpoint_util.client import Client as EndpointUtilClient
from alibabacloud_tea_util import models as util_models
from alibabacloud_openapi_util.client import Client as OpenApiUtilClient

from pai.libs.alibabacloud_paiflow20210202 import models as paiflow_20210202_models


class Client(OpenApiClient):
    """
    *\
    """
    def __init__(
        self,
        config: open_api_models.Config,
    ):
        super().__init__(config)
        self._endpoint_rule = 'central'
        self._endpoint_map = {
            'cn-beijing': 'pai.cn-beijing.aliyuncs.com',
            'cn-hangzhou': 'pai.cn-hangzhou.data.aliyun.com',
            'cn-shanghai': 'pai.cn-shanghai.aliyuncs.com',
            'cn-shenzhen': 'pai.cn-shenzhen.aliyuncs.com',
            'cn-hongkong': 'pai.cn-hongkong.aliyuncs.com',
            'ap-southeast-1': 'pai.ap-southeast-1.aliyuncs.com',
            'ap-southeast-2': 'pai.ap-southeast-2.aliyuncs.com',
            'ap-southeast-3': 'pai.ap-southeast-3.aliyuncs.com',
            'ap-southeast-5': 'pai.ap-southeast-5.aliyuncs.com',
            'us-west-1': 'pai.us-west-1.aliyuncs.com',
            'us-east-1': 'pai.us-east-1.aliyuncs.com',
            'eu-central-1': 'pai.eu-central-1.aliyuncs.com',
            'me-east-1': 'pai.me-east-1.aliyuncs.com',
            'ap-south-1': 'pai.ap-south-1.aliyuncs.com'
        }
        self.check_config(config)
        self._endpoint = self.get_endpoint('paiflow', self._region_id, self._endpoint_rule, self._network, self._suffix, self._endpoint_map, self._endpoint)

    def get_endpoint(
        self,
        product_id: str,
        region_id: str,
        endpoint_rule: str,
        network: str,
        suffix: str,
        endpoint_map: Dict[str, str],
        endpoint: str,
    ) -> str:
        if not UtilClient.empty(endpoint):
            return endpoint
        if not UtilClient.is_unset(endpoint_map) and not UtilClient.empty(endpoint_map.get(region_id)):
            return endpoint_map.get(region_id)
        return EndpointUtilClient.get_endpoint_rules(product_id, region_id, endpoint_rule, network, suffix)

    def create_pipeline(
        self,
        request: paiflow_20210202_models.CreatePipelineRequest,
    ) -> paiflow_20210202_models.CreatePipelineResponse:
        runtime = util_models.RuntimeOptions()
        headers = {}
        return self.create_pipeline_with_options(request, headers, runtime)

    async def create_pipeline_async(
        self,
        request: paiflow_20210202_models.CreatePipelineRequest,
    ) -> paiflow_20210202_models.CreatePipelineResponse:
        runtime = util_models.RuntimeOptions()
        headers = {}
        return await self.create_pipeline_with_options_async(request, headers, runtime)

    def create_pipeline_with_options(
        self,
        request: paiflow_20210202_models.CreatePipelineRequest,
        headers: Dict[str, str],
        runtime: util_models.RuntimeOptions,
    ) -> paiflow_20210202_models.CreatePipelineResponse:
        UtilClient.validate_model(request)
        body = {}
        if not UtilClient.is_unset(request.manifest):
            body['Manifest'] = request.manifest
        if not UtilClient.is_unset(request.workspace_id):
            body['WorkspaceId'] = request.workspace_id
        req = open_api_models.OpenApiRequest(
            headers=headers,
            body=OpenApiUtilClient.parse_to_map(body)
        )
        params = open_api_models.Params(
            action='CreatePipeline',
            version='2021-02-02',
            protocol='HTTPS',
            pathname=f'/api/v1/pipelines',
            method='POST',
            auth_type='AK',
            style='ROA',
            req_body_type='json',
            body_type='json'
        )
        return TeaCore.from_map(
            paiflow_20210202_models.CreatePipelineResponse(),
            self.call_api(params, req, runtime)
        )

    async def create_pipeline_with_options_async(
        self,
        request: paiflow_20210202_models.CreatePipelineRequest,
        headers: Dict[str, str],
        runtime: util_models.RuntimeOptions,
    ) -> paiflow_20210202_models.CreatePipelineResponse:
        UtilClient.validate_model(request)
        body = {}
        if not UtilClient.is_unset(request.manifest):
            body['Manifest'] = request.manifest
        if not UtilClient.is_unset(request.workspace_id):
            body['WorkspaceId'] = request.workspace_id
        req = open_api_models.OpenApiRequest(
            headers=headers,
            body=OpenApiUtilClient.parse_to_map(body)
        )
        params = open_api_models.Params(
            action='CreatePipeline',
            version='2021-02-02',
            protocol='HTTPS',
            pathname=f'/api/v1/pipelines',
            method='POST',
            auth_type='AK',
            style='ROA',
            req_body_type='json',
            body_type='json'
        )
        return TeaCore.from_map(
            paiflow_20210202_models.CreatePipelineResponse(),
            await self.call_api_async(params, req, runtime)
        )

    def create_pipeline_release(
        self,
        pipeline_id: str,
        request: paiflow_20210202_models.CreatePipelineReleaseRequest,
    ) -> paiflow_20210202_models.CreatePipelineReleaseResponse:
        runtime = util_models.RuntimeOptions()
        headers = {}
        return self.create_pipeline_release_with_options(pipeline_id, request, headers, runtime)

    async def create_pipeline_release_async(
        self,
        pipeline_id: str,
        request: paiflow_20210202_models.CreatePipelineReleaseRequest,
    ) -> paiflow_20210202_models.CreatePipelineReleaseResponse:
        runtime = util_models.RuntimeOptions()
        headers = {}
        return await self.create_pipeline_release_with_options_async(pipeline_id, request, headers, runtime)

    def create_pipeline_release_with_options(
        self,
        pipeline_id: str,
        request: paiflow_20210202_models.CreatePipelineReleaseRequest,
        headers: Dict[str, str],
        runtime: util_models.RuntimeOptions,
    ) -> paiflow_20210202_models.CreatePipelineReleaseResponse:
        UtilClient.validate_model(request)
        pipeline_id = OpenApiUtilClient.get_encode_param(pipeline_id)
        body = {}
        if not UtilClient.is_unset(request.target_pipeline_provider):
            body['TargetPipelineProvider'] = request.target_pipeline_provider
        req = open_api_models.OpenApiRequest(
            headers=headers,
            body=OpenApiUtilClient.parse_to_map(body)
        )
        params = open_api_models.Params(
            action='CreatePipelineRelease',
            version='2021-02-02',
            protocol='HTTPS',
            pathname=f'/api/v1/pipelines/{pipeline_id}/releases',
            method='PUT',
            auth_type='AK',
            style='ROA',
            req_body_type='json',
            body_type='json'
        )
        return TeaCore.from_map(
            paiflow_20210202_models.CreatePipelineReleaseResponse(),
            self.call_api(params, req, runtime)
        )

    async def create_pipeline_release_with_options_async(
        self,
        pipeline_id: str,
        request: paiflow_20210202_models.CreatePipelineReleaseRequest,
        headers: Dict[str, str],
        runtime: util_models.RuntimeOptions,
    ) -> paiflow_20210202_models.CreatePipelineReleaseResponse:
        UtilClient.validate_model(request)
        pipeline_id = OpenApiUtilClient.get_encode_param(pipeline_id)
        body = {}
        if not UtilClient.is_unset(request.target_pipeline_provider):
            body['TargetPipelineProvider'] = request.target_pipeline_provider
        req = open_api_models.OpenApiRequest(
            headers=headers,
            body=OpenApiUtilClient.parse_to_map(body)
        )
        params = open_api_models.Params(
            action='CreatePipelineRelease',
            version='2021-02-02',
            protocol='HTTPS',
            pathname=f'/api/v1/pipelines/{pipeline_id}/releases',
            method='PUT',
            auth_type='AK',
            style='ROA',
            req_body_type='json',
            body_type='json'
        )
        return TeaCore.from_map(
            paiflow_20210202_models.CreatePipelineReleaseResponse(),
            await self.call_api_async(params, req, runtime)
        )

    def create_run(
        self,
        request: paiflow_20210202_models.CreateRunRequest,
    ) -> paiflow_20210202_models.CreateRunResponse:
        runtime = util_models.RuntimeOptions()
        headers = {}
        return self.create_run_with_options(request, headers, runtime)

    async def create_run_async(
        self,
        request: paiflow_20210202_models.CreateRunRequest,
    ) -> paiflow_20210202_models.CreateRunResponse:
        runtime = util_models.RuntimeOptions()
        headers = {}
        return await self.create_run_with_options_async(request, headers, runtime)

    def create_run_with_options(
        self,
        request: paiflow_20210202_models.CreateRunRequest,
        headers: Dict[str, str],
        runtime: util_models.RuntimeOptions,
    ) -> paiflow_20210202_models.CreateRunResponse:
        UtilClient.validate_model(request)
        body = {}
        if not UtilClient.is_unset(request.accessibility):
            body['Accessibility'] = request.accessibility
        if not UtilClient.is_unset(request.arguments):
            body['Arguments'] = request.arguments
        if not UtilClient.is_unset(request.experiment_id):
            body['ExperimentId'] = request.experiment_id
        if not UtilClient.is_unset(request.name):
            body['Name'] = request.name
        if not UtilClient.is_unset(request.no_confirm_required):
            body['NoConfirmRequired'] = request.no_confirm_required
        if not UtilClient.is_unset(request.options):
            body['Options'] = request.options
        if not UtilClient.is_unset(request.pipeline_id):
            body['PipelineId'] = request.pipeline_id
        if not UtilClient.is_unset(request.pipeline_manifest):
            body['PipelineManifest'] = request.pipeline_manifest
        if not UtilClient.is_unset(request.source):
            body['Source'] = request.source
        if not UtilClient.is_unset(request.workspace_id):
            body['WorkspaceId'] = request.workspace_id
        req = open_api_models.OpenApiRequest(
            headers=headers,
            body=OpenApiUtilClient.parse_to_map(body)
        )
        params = open_api_models.Params(
            action='CreateRun',
            version='2021-02-02',
            protocol='HTTPS',
            pathname=f'/api/v1/runs',
            method='POST',
            auth_type='AK',
            style='ROA',
            req_body_type='json',
            body_type='json'
        )
        return TeaCore.from_map(
            paiflow_20210202_models.CreateRunResponse(),
            self.call_api(params, req, runtime)
        )

    async def create_run_with_options_async(
        self,
        request: paiflow_20210202_models.CreateRunRequest,
        headers: Dict[str, str],
        runtime: util_models.RuntimeOptions,
    ) -> paiflow_20210202_models.CreateRunResponse:
        UtilClient.validate_model(request)
        body = {}
        if not UtilClient.is_unset(request.accessibility):
            body['Accessibility'] = request.accessibility
        if not UtilClient.is_unset(request.arguments):
            body['Arguments'] = request.arguments
        if not UtilClient.is_unset(request.experiment_id):
            body['ExperimentId'] = request.experiment_id
        if not UtilClient.is_unset(request.name):
            body['Name'] = request.name
        if not UtilClient.is_unset(request.no_confirm_required):
            body['NoConfirmRequired'] = request.no_confirm_required
        if not UtilClient.is_unset(request.options):
            body['Options'] = request.options
        if not UtilClient.is_unset(request.pipeline_id):
            body['PipelineId'] = request.pipeline_id
        if not UtilClient.is_unset(request.pipeline_manifest):
            body['PipelineManifest'] = request.pipeline_manifest
        if not UtilClient.is_unset(request.source):
            body['Source'] = request.source
        if not UtilClient.is_unset(request.workspace_id):
            body['WorkspaceId'] = request.workspace_id
        req = open_api_models.OpenApiRequest(
            headers=headers,
            body=OpenApiUtilClient.parse_to_map(body)
        )
        params = open_api_models.Params(
            action='CreateRun',
            version='2021-02-02',
            protocol='HTTPS',
            pathname=f'/api/v1/runs',
            method='POST',
            auth_type='AK',
            style='ROA',
            req_body_type='json',
            body_type='json'
        )
        return TeaCore.from_map(
            paiflow_20210202_models.CreateRunResponse(),
            await self.call_api_async(params, req, runtime)
        )

    def delete_pipeline(
        self,
        pipeline_id: str,
    ) -> paiflow_20210202_models.DeletePipelineResponse:
        runtime = util_models.RuntimeOptions()
        headers = {}
        return self.delete_pipeline_with_options(pipeline_id, headers, runtime)

    async def delete_pipeline_async(
        self,
        pipeline_id: str,
    ) -> paiflow_20210202_models.DeletePipelineResponse:
        runtime = util_models.RuntimeOptions()
        headers = {}
        return await self.delete_pipeline_with_options_async(pipeline_id, headers, runtime)

    def delete_pipeline_with_options(
        self,
        pipeline_id: str,
        headers: Dict[str, str],
        runtime: util_models.RuntimeOptions,
    ) -> paiflow_20210202_models.DeletePipelineResponse:
        pipeline_id = OpenApiUtilClient.get_encode_param(pipeline_id)
        req = open_api_models.OpenApiRequest(
            headers=headers
        )
        params = open_api_models.Params(
            action='DeletePipeline',
            version='2021-02-02',
            protocol='HTTPS',
            pathname=f'/api/v1/pipelines/{pipeline_id}',
            method='DELETE',
            auth_type='AK',
            style='ROA',
            req_body_type='json',
            body_type='json'
        )
        return TeaCore.from_map(
            paiflow_20210202_models.DeletePipelineResponse(),
            self.call_api(params, req, runtime)
        )

    async def delete_pipeline_with_options_async(
        self,
        pipeline_id: str,
        headers: Dict[str, str],
        runtime: util_models.RuntimeOptions,
    ) -> paiflow_20210202_models.DeletePipelineResponse:
        pipeline_id = OpenApiUtilClient.get_encode_param(pipeline_id)
        req = open_api_models.OpenApiRequest(
            headers=headers
        )
        params = open_api_models.Params(
            action='DeletePipeline',
            version='2021-02-02',
            protocol='HTTPS',
            pathname=f'/api/v1/pipelines/{pipeline_id}',
            method='DELETE',
            auth_type='AK',
            style='ROA',
            req_body_type='json',
            body_type='json'
        )
        return TeaCore.from_map(
            paiflow_20210202_models.DeletePipelineResponse(),
            await self.call_api_async(params, req, runtime)
        )

    def delete_run(
        self,
        run_id: str,
    ) -> paiflow_20210202_models.DeleteRunResponse:
        runtime = util_models.RuntimeOptions()
        headers = {}
        return self.delete_run_with_options(run_id, headers, runtime)

    async def delete_run_async(
        self,
        run_id: str,
    ) -> paiflow_20210202_models.DeleteRunResponse:
        runtime = util_models.RuntimeOptions()
        headers = {}
        return await self.delete_run_with_options_async(run_id, headers, runtime)

    def delete_run_with_options(
        self,
        run_id: str,
        headers: Dict[str, str],
        runtime: util_models.RuntimeOptions,
    ) -> paiflow_20210202_models.DeleteRunResponse:
        run_id = OpenApiUtilClient.get_encode_param(run_id)
        req = open_api_models.OpenApiRequest(
            headers=headers
        )
        params = open_api_models.Params(
            action='DeleteRun',
            version='2021-02-02',
            protocol='HTTPS',
            pathname=f'/api/v1/runs/{run_id}',
            method='DELETE',
            auth_type='AK',
            style='ROA',
            req_body_type='json',
            body_type='json'
        )
        return TeaCore.from_map(
            paiflow_20210202_models.DeleteRunResponse(),
            self.call_api(params, req, runtime)
        )

    async def delete_run_with_options_async(
        self,
        run_id: str,
        headers: Dict[str, str],
        runtime: util_models.RuntimeOptions,
    ) -> paiflow_20210202_models.DeleteRunResponse:
        run_id = OpenApiUtilClient.get_encode_param(run_id)
        req = open_api_models.OpenApiRequest(
            headers=headers
        )
        params = open_api_models.Params(
            action='DeleteRun',
            version='2021-02-02',
            protocol='HTTPS',
            pathname=f'/api/v1/runs/{run_id}',
            method='DELETE',
            auth_type='AK',
            style='ROA',
            req_body_type='json',
            body_type='json'
        )
        return TeaCore.from_map(
            paiflow_20210202_models.DeleteRunResponse(),
            await self.call_api_async(params, req, runtime)
        )

    def get_caller_provider(self) -> paiflow_20210202_models.GetCallerProviderResponse:
        runtime = util_models.RuntimeOptions()
        headers = {}
        return self.get_caller_provider_with_options(headers, runtime)

    async def get_caller_provider_async(self) -> paiflow_20210202_models.GetCallerProviderResponse:
        runtime = util_models.RuntimeOptions()
        headers = {}
        return await self.get_caller_provider_with_options_async(headers, runtime)

    def get_caller_provider_with_options(
        self,
        headers: Dict[str, str],
        runtime: util_models.RuntimeOptions,
    ) -> paiflow_20210202_models.GetCallerProviderResponse:
        req = open_api_models.OpenApiRequest(
            headers=headers
        )
        params = open_api_models.Params(
            action='GetCallerProvider',
            version='2021-02-02',
            protocol='HTTPS',
            pathname=f'/api/v1/provider',
            method='GET',
            auth_type='AK',
            style='ROA',
            req_body_type='json',
            body_type='json'
        )
        return TeaCore.from_map(
            paiflow_20210202_models.GetCallerProviderResponse(),
            self.call_api(params, req, runtime)
        )

    async def get_caller_provider_with_options_async(
        self,
        headers: Dict[str, str],
        runtime: util_models.RuntimeOptions,
    ) -> paiflow_20210202_models.GetCallerProviderResponse:
        req = open_api_models.OpenApiRequest(
            headers=headers
        )
        params = open_api_models.Params(
            action='GetCallerProvider',
            version='2021-02-02',
            protocol='HTTPS',
            pathname=f'/api/v1/provider',
            method='GET',
            auth_type='AK',
            style='ROA',
            req_body_type='json',
            body_type='json'
        )
        return TeaCore.from_map(
            paiflow_20210202_models.GetCallerProviderResponse(),
            await self.call_api_async(params, req, runtime)
        )

    def get_node(
        self,
        run_id: str,
        node_id: str,
        request: paiflow_20210202_models.GetNodeRequest,
    ) -> paiflow_20210202_models.GetNodeResponse:
        runtime = util_models.RuntimeOptions()
        headers = {}
        return self.get_node_with_options(run_id, node_id, request, headers, runtime)

    async def get_node_async(
        self,
        run_id: str,
        node_id: str,
        request: paiflow_20210202_models.GetNodeRequest,
    ) -> paiflow_20210202_models.GetNodeResponse:
        runtime = util_models.RuntimeOptions()
        headers = {}
        return await self.get_node_with_options_async(run_id, node_id, request, headers, runtime)

    def get_node_with_options(
        self,
        run_id: str,
        node_id: str,
        request: paiflow_20210202_models.GetNodeRequest,
        headers: Dict[str, str],
        runtime: util_models.RuntimeOptions,
    ) -> paiflow_20210202_models.GetNodeResponse:
        UtilClient.validate_model(request)
        run_id = OpenApiUtilClient.get_encode_param(run_id)
        node_id = OpenApiUtilClient.get_encode_param(node_id)
        query = {}
        if not UtilClient.is_unset(request.depth):
            query['Depth'] = request.depth
        req = open_api_models.OpenApiRequest(
            headers=headers,
            query=OpenApiUtilClient.query(query)
        )
        params = open_api_models.Params(
            action='GetNode',
            version='2021-02-02',
            protocol='HTTPS',
            pathname=f'/api/v1/runs/{run_id}/nodes/{node_id}',
            method='GET',
            auth_type='AK',
            style='ROA',
            req_body_type='json',
            body_type='json'
        )
        return TeaCore.from_map(
            paiflow_20210202_models.GetNodeResponse(),
            self.call_api(params, req, runtime)
        )

    async def get_node_with_options_async(
        self,
        run_id: str,
        node_id: str,
        request: paiflow_20210202_models.GetNodeRequest,
        headers: Dict[str, str],
        runtime: util_models.RuntimeOptions,
    ) -> paiflow_20210202_models.GetNodeResponse:
        UtilClient.validate_model(request)
        run_id = OpenApiUtilClient.get_encode_param(run_id)
        node_id = OpenApiUtilClient.get_encode_param(node_id)
        query = {}
        if not UtilClient.is_unset(request.depth):
            query['Depth'] = request.depth
        req = open_api_models.OpenApiRequest(
            headers=headers,
            query=OpenApiUtilClient.query(query)
        )
        params = open_api_models.Params(
            action='GetNode',
            version='2021-02-02',
            protocol='HTTPS',
            pathname=f'/api/v1/runs/{run_id}/nodes/{node_id}',
            method='GET',
            auth_type='AK',
            style='ROA',
            req_body_type='json',
            body_type='json'
        )
        return TeaCore.from_map(
            paiflow_20210202_models.GetNodeResponse(),
            await self.call_api_async(params, req, runtime)
        )

    def get_pipeline(
        self,
        pipeline_id: str,
    ) -> paiflow_20210202_models.GetPipelineResponse:
        runtime = util_models.RuntimeOptions()
        headers = {}
        return self.get_pipeline_with_options(pipeline_id, headers, runtime)

    async def get_pipeline_async(
        self,
        pipeline_id: str,
    ) -> paiflow_20210202_models.GetPipelineResponse:
        runtime = util_models.RuntimeOptions()
        headers = {}
        return await self.get_pipeline_with_options_async(pipeline_id, headers, runtime)

    def get_pipeline_with_options(
        self,
        pipeline_id: str,
        headers: Dict[str, str],
        runtime: util_models.RuntimeOptions,
    ) -> paiflow_20210202_models.GetPipelineResponse:
        pipeline_id = OpenApiUtilClient.get_encode_param(pipeline_id)
        req = open_api_models.OpenApiRequest(
            headers=headers
        )
        params = open_api_models.Params(
            action='GetPipeline',
            version='2021-02-02',
            protocol='HTTPS',
            pathname=f'/api/v1/pipelines/{pipeline_id}',
            method='GET',
            auth_type='AK',
            style='ROA',
            req_body_type='json',
            body_type='json'
        )
        return TeaCore.from_map(
            paiflow_20210202_models.GetPipelineResponse(),
            self.call_api(params, req, runtime)
        )

    async def get_pipeline_with_options_async(
        self,
        pipeline_id: str,
        headers: Dict[str, str],
        runtime: util_models.RuntimeOptions,
    ) -> paiflow_20210202_models.GetPipelineResponse:
        pipeline_id = OpenApiUtilClient.get_encode_param(pipeline_id)
        req = open_api_models.OpenApiRequest(
            headers=headers
        )
        params = open_api_models.Params(
            action='GetPipeline',
            version='2021-02-02',
            protocol='HTTPS',
            pathname=f'/api/v1/pipelines/{pipeline_id}',
            method='GET',
            auth_type='AK',
            style='ROA',
            req_body_type='json',
            body_type='json'
        )
        return TeaCore.from_map(
            paiflow_20210202_models.GetPipelineResponse(),
            await self.call_api_async(params, req, runtime)
        )

    def get_pipeline_schema(
        self,
        pipeline_id: str,
    ) -> paiflow_20210202_models.GetPipelineSchemaResponse:
        runtime = util_models.RuntimeOptions()
        headers = {}
        return self.get_pipeline_schema_with_options(pipeline_id, headers, runtime)

    async def get_pipeline_schema_async(
        self,
        pipeline_id: str,
    ) -> paiflow_20210202_models.GetPipelineSchemaResponse:
        runtime = util_models.RuntimeOptions()
        headers = {}
        return await self.get_pipeline_schema_with_options_async(pipeline_id, headers, runtime)

    def get_pipeline_schema_with_options(
        self,
        pipeline_id: str,
        headers: Dict[str, str],
        runtime: util_models.RuntimeOptions,
    ) -> paiflow_20210202_models.GetPipelineSchemaResponse:
        pipeline_id = OpenApiUtilClient.get_encode_param(pipeline_id)
        req = open_api_models.OpenApiRequest(
            headers=headers
        )
        params = open_api_models.Params(
            action='GetPipelineSchema',
            version='2021-02-02',
            protocol='HTTPS',
            pathname=f'/api/v1/pipelines/{pipeline_id}/schema',
            method='GET',
            auth_type='AK',
            style='ROA',
            req_body_type='json',
            body_type='json'
        )
        return TeaCore.from_map(
            paiflow_20210202_models.GetPipelineSchemaResponse(),
            self.call_api(params, req, runtime)
        )

    async def get_pipeline_schema_with_options_async(
        self,
        pipeline_id: str,
        headers: Dict[str, str],
        runtime: util_models.RuntimeOptions,
    ) -> paiflow_20210202_models.GetPipelineSchemaResponse:
        pipeline_id = OpenApiUtilClient.get_encode_param(pipeline_id)
        req = open_api_models.OpenApiRequest(
            headers=headers
        )
        params = open_api_models.Params(
            action='GetPipelineSchema',
            version='2021-02-02',
            protocol='HTTPS',
            pathname=f'/api/v1/pipelines/{pipeline_id}/schema',
            method='GET',
            auth_type='AK',
            style='ROA',
            req_body_type='json',
            body_type='json'
        )
        return TeaCore.from_map(
            paiflow_20210202_models.GetPipelineSchemaResponse(),
            await self.call_api_async(params, req, runtime)
        )

    def get_run(
        self,
        run_id: str,
        request: paiflow_20210202_models.GetRunRequest,
    ) -> paiflow_20210202_models.GetRunResponse:
        runtime = util_models.RuntimeOptions()
        headers = {}
        return self.get_run_with_options(run_id, request, headers, runtime)

    async def get_run_async(
        self,
        run_id: str,
        request: paiflow_20210202_models.GetRunRequest,
    ) -> paiflow_20210202_models.GetRunResponse:
        runtime = util_models.RuntimeOptions()
        headers = {}
        return await self.get_run_with_options_async(run_id, request, headers, runtime)

    def get_run_with_options(
        self,
        run_id: str,
        request: paiflow_20210202_models.GetRunRequest,
        headers: Dict[str, str],
        runtime: util_models.RuntimeOptions,
    ) -> paiflow_20210202_models.GetRunResponse:
        UtilClient.validate_model(request)
        run_id = OpenApiUtilClient.get_encode_param(run_id)
        query = {}
        if not UtilClient.is_unset(request.manifest_type):
            query['ManifestType'] = request.manifest_type
        if not UtilClient.is_unset(request.verbose):
            query['Verbose'] = request.verbose
        req = open_api_models.OpenApiRequest(
            headers=headers,
            query=OpenApiUtilClient.query(query)
        )
        params = open_api_models.Params(
            action='GetRun',
            version='2021-02-02',
            protocol='HTTPS',
            pathname=f'/api/v1/runs/{run_id}',
            method='GET',
            auth_type='AK',
            style='ROA',
            req_body_type='json',
            body_type='json'
        )
        return TeaCore.from_map(
            paiflow_20210202_models.GetRunResponse(),
            self.call_api(params, req, runtime)
        )

    async def get_run_with_options_async(
        self,
        run_id: str,
        request: paiflow_20210202_models.GetRunRequest,
        headers: Dict[str, str],
        runtime: util_models.RuntimeOptions,
    ) -> paiflow_20210202_models.GetRunResponse:
        UtilClient.validate_model(request)
        run_id = OpenApiUtilClient.get_encode_param(run_id)
        query = {}
        if not UtilClient.is_unset(request.manifest_type):
            query['ManifestType'] = request.manifest_type
        if not UtilClient.is_unset(request.verbose):
            query['Verbose'] = request.verbose
        req = open_api_models.OpenApiRequest(
            headers=headers,
            query=OpenApiUtilClient.query(query)
        )
        params = open_api_models.Params(
            action='GetRun',
            version='2021-02-02',
            protocol='HTTPS',
            pathname=f'/api/v1/runs/{run_id}',
            method='GET',
            auth_type='AK',
            style='ROA',
            req_body_type='json',
            body_type='json'
        )
        return TeaCore.from_map(
            paiflow_20210202_models.GetRunResponse(),
            await self.call_api_async(params, req, runtime)
        )

    def get_run_statistics(
        self,
        request: paiflow_20210202_models.GetRunStatisticsRequest,
    ) -> paiflow_20210202_models.GetRunStatisticsResponse:
        runtime = util_models.RuntimeOptions()
        headers = {}
        return self.get_run_statistics_with_options(request, headers, runtime)

    async def get_run_statistics_async(
        self,
        request: paiflow_20210202_models.GetRunStatisticsRequest,
    ) -> paiflow_20210202_models.GetRunStatisticsResponse:
        runtime = util_models.RuntimeOptions()
        headers = {}
        return await self.get_run_statistics_with_options_async(request, headers, runtime)

    def get_run_statistics_with_options(
        self,
        tmp_req: paiflow_20210202_models.GetRunStatisticsRequest,
        headers: Dict[str, str],
        runtime: util_models.RuntimeOptions,
    ) -> paiflow_20210202_models.GetRunStatisticsResponse:
        UtilClient.validate_model(tmp_req)
        request = paiflow_20210202_models.GetRunStatisticsShrinkRequest()
        OpenApiUtilClient.convert(tmp_req, request)
        if not UtilClient.is_unset(tmp_req.status):
            request.status_shrink = OpenApiUtilClient.array_to_string_with_specified_style(tmp_req.status, 'Status', 'json')
        query = {}
        if not UtilClient.is_unset(request.experiment_id):
            query['ExperimentId'] = request.experiment_id
        if not UtilClient.is_unset(request.is_show_all):
            query['IsShowAll'] = request.is_show_all
        if not UtilClient.is_unset(request.name):
            query['Name'] = request.name
        if not UtilClient.is_unset(request.pipeline_id):
            query['PipelineId'] = request.pipeline_id
        if not UtilClient.is_unset(request.scope):
            query['Scope'] = request.scope
        if not UtilClient.is_unset(request.source):
            query['Source'] = request.source
        if not UtilClient.is_unset(request.status_shrink):
            query['Status'] = request.status_shrink
        if not UtilClient.is_unset(request.workspace_id):
            query['WorkspaceId'] = request.workspace_id
        req = open_api_models.OpenApiRequest(
            headers=headers,
            query=OpenApiUtilClient.query(query)
        )
        params = open_api_models.Params(
            action='GetRunStatistics',
            version='2021-02-02',
            protocol='HTTPS',
            pathname=f'/api/v1/statistics/runs',
            method='GET',
            auth_type='AK',
            style='ROA',
            req_body_type='json',
            body_type='json'
        )
        return TeaCore.from_map(
            paiflow_20210202_models.GetRunStatisticsResponse(),
            self.call_api(params, req, runtime)
        )

    async def get_run_statistics_with_options_async(
        self,
        tmp_req: paiflow_20210202_models.GetRunStatisticsRequest,
        headers: Dict[str, str],
        runtime: util_models.RuntimeOptions,
    ) -> paiflow_20210202_models.GetRunStatisticsResponse:
        UtilClient.validate_model(tmp_req)
        request = paiflow_20210202_models.GetRunStatisticsShrinkRequest()
        OpenApiUtilClient.convert(tmp_req, request)
        if not UtilClient.is_unset(tmp_req.status):
            request.status_shrink = OpenApiUtilClient.array_to_string_with_specified_style(tmp_req.status, 'Status', 'json')
        query = {}
        if not UtilClient.is_unset(request.experiment_id):
            query['ExperimentId'] = request.experiment_id
        if not UtilClient.is_unset(request.is_show_all):
            query['IsShowAll'] = request.is_show_all
        if not UtilClient.is_unset(request.name):
            query['Name'] = request.name
        if not UtilClient.is_unset(request.pipeline_id):
            query['PipelineId'] = request.pipeline_id
        if not UtilClient.is_unset(request.scope):
            query['Scope'] = request.scope
        if not UtilClient.is_unset(request.source):
            query['Source'] = request.source
        if not UtilClient.is_unset(request.status_shrink):
            query['Status'] = request.status_shrink
        if not UtilClient.is_unset(request.workspace_id):
            query['WorkspaceId'] = request.workspace_id
        req = open_api_models.OpenApiRequest(
            headers=headers,
            query=OpenApiUtilClient.query(query)
        )
        params = open_api_models.Params(
            action='GetRunStatistics',
            version='2021-02-02',
            protocol='HTTPS',
            pathname=f'/api/v1/statistics/runs',
            method='GET',
            auth_type='AK',
            style='ROA',
            req_body_type='json',
            body_type='json'
        )
        return TeaCore.from_map(
            paiflow_20210202_models.GetRunStatisticsResponse(),
            await self.call_api_async(params, req, runtime)
        )

    def list_node_logs(
        self,
        run_id: str,
        node_id: str,
        request: paiflow_20210202_models.ListNodeLogsRequest,
    ) -> paiflow_20210202_models.ListNodeLogsResponse:
        runtime = util_models.RuntimeOptions()
        headers = {}
        return self.list_node_logs_with_options(run_id, node_id, request, headers, runtime)

    async def list_node_logs_async(
        self,
        run_id: str,
        node_id: str,
        request: paiflow_20210202_models.ListNodeLogsRequest,
    ) -> paiflow_20210202_models.ListNodeLogsResponse:
        runtime = util_models.RuntimeOptions()
        headers = {}
        return await self.list_node_logs_with_options_async(run_id, node_id, request, headers, runtime)

    def list_node_logs_with_options(
        self,
        run_id: str,
        node_id: str,
        request: paiflow_20210202_models.ListNodeLogsRequest,
        headers: Dict[str, str],
        runtime: util_models.RuntimeOptions,
    ) -> paiflow_20210202_models.ListNodeLogsResponse:
        UtilClient.validate_model(request)
        run_id = OpenApiUtilClient.get_encode_param(run_id)
        node_id = OpenApiUtilClient.get_encode_param(node_id)
        query = {}
        if not UtilClient.is_unset(request.from_time_in_seconds):
            query['FromTimeInSeconds'] = request.from_time_in_seconds
        if not UtilClient.is_unset(request.keyword):
            query['Keyword'] = request.keyword
        if not UtilClient.is_unset(request.offset):
            query['Offset'] = request.offset
        if not UtilClient.is_unset(request.page_size):
            query['PageSize'] = request.page_size
        if not UtilClient.is_unset(request.reverse):
            query['Reverse'] = request.reverse
        if not UtilClient.is_unset(request.to_time_in_seconds):
            query['ToTimeInSeconds'] = request.to_time_in_seconds
        req = open_api_models.OpenApiRequest(
            headers=headers,
            query=OpenApiUtilClient.query(query)
        )
        params = open_api_models.Params(
            action='ListNodeLogs',
            version='2021-02-02',
            protocol='HTTPS',
            pathname=f'/api/v1/runs/{run_id}/nodes/{node_id}/logs',
            method='GET',
            auth_type='AK',
            style='ROA',
            req_body_type='json',
            body_type='json'
        )
        return TeaCore.from_map(
            paiflow_20210202_models.ListNodeLogsResponse(),
            self.call_api(params, req, runtime)
        )

    async def list_node_logs_with_options_async(
        self,
        run_id: str,
        node_id: str,
        request: paiflow_20210202_models.ListNodeLogsRequest,
        headers: Dict[str, str],
        runtime: util_models.RuntimeOptions,
    ) -> paiflow_20210202_models.ListNodeLogsResponse:
        UtilClient.validate_model(request)
        run_id = OpenApiUtilClient.get_encode_param(run_id)
        node_id = OpenApiUtilClient.get_encode_param(node_id)
        query = {}
        if not UtilClient.is_unset(request.from_time_in_seconds):
            query['FromTimeInSeconds'] = request.from_time_in_seconds
        if not UtilClient.is_unset(request.keyword):
            query['Keyword'] = request.keyword
        if not UtilClient.is_unset(request.offset):
            query['Offset'] = request.offset
        if not UtilClient.is_unset(request.page_size):
            query['PageSize'] = request.page_size
        if not UtilClient.is_unset(request.reverse):
            query['Reverse'] = request.reverse
        if not UtilClient.is_unset(request.to_time_in_seconds):
            query['ToTimeInSeconds'] = request.to_time_in_seconds
        req = open_api_models.OpenApiRequest(
            headers=headers,
            query=OpenApiUtilClient.query(query)
        )
        params = open_api_models.Params(
            action='ListNodeLogs',
            version='2021-02-02',
            protocol='HTTPS',
            pathname=f'/api/v1/runs/{run_id}/nodes/{node_id}/logs',
            method='GET',
            auth_type='AK',
            style='ROA',
            req_body_type='json',
            body_type='json'
        )
        return TeaCore.from_map(
            paiflow_20210202_models.ListNodeLogsResponse(),
            await self.call_api_async(params, req, runtime)
        )

    def list_node_outputs(
        self,
        run_id: str,
        node_id: str,
        request: paiflow_20210202_models.ListNodeOutputsRequest,
    ) -> paiflow_20210202_models.ListNodeOutputsResponse:
        runtime = util_models.RuntimeOptions()
        headers = {}
        return self.list_node_outputs_with_options(run_id, node_id, request, headers, runtime)

    async def list_node_outputs_async(
        self,
        run_id: str,
        node_id: str,
        request: paiflow_20210202_models.ListNodeOutputsRequest,
    ) -> paiflow_20210202_models.ListNodeOutputsResponse:
        runtime = util_models.RuntimeOptions()
        headers = {}
        return await self.list_node_outputs_with_options_async(run_id, node_id, request, headers, runtime)

    def list_node_outputs_with_options(
        self,
        run_id: str,
        node_id: str,
        request: paiflow_20210202_models.ListNodeOutputsRequest,
        headers: Dict[str, str],
        runtime: util_models.RuntimeOptions,
    ) -> paiflow_20210202_models.ListNodeOutputsResponse:
        UtilClient.validate_model(request)
        run_id = OpenApiUtilClient.get_encode_param(run_id)
        node_id = OpenApiUtilClient.get_encode_param(node_id)
        query = {}
        if not UtilClient.is_unset(request.depth):
            query['Depth'] = request.depth
        if not UtilClient.is_unset(request.name):
            query['Name'] = request.name
        if not UtilClient.is_unset(request.order):
            query['Order'] = request.order
        if not UtilClient.is_unset(request.page_number):
            query['PageNumber'] = request.page_number
        if not UtilClient.is_unset(request.page_size):
            query['PageSize'] = request.page_size
        if not UtilClient.is_unset(request.sort_by):
            query['SortBy'] = request.sort_by
        if not UtilClient.is_unset(request.type):
            query['Type'] = request.type
        req = open_api_models.OpenApiRequest(
            headers=headers,
            query=OpenApiUtilClient.query(query)
        )
        params = open_api_models.Params(
            action='ListNodeOutputs',
            version='2021-02-02',
            protocol='HTTPS',
            pathname=f'/api/v1/runs/{run_id}/nodes/{node_id}/outputs',
            method='GET',
            auth_type='AK',
            style='ROA',
            req_body_type='json',
            body_type='json'
        )
        return TeaCore.from_map(
            paiflow_20210202_models.ListNodeOutputsResponse(),
            self.call_api(params, req, runtime)
        )

    async def list_node_outputs_with_options_async(
        self,
        run_id: str,
        node_id: str,
        request: paiflow_20210202_models.ListNodeOutputsRequest,
        headers: Dict[str, str],
        runtime: util_models.RuntimeOptions,
    ) -> paiflow_20210202_models.ListNodeOutputsResponse:
        UtilClient.validate_model(request)
        run_id = OpenApiUtilClient.get_encode_param(run_id)
        node_id = OpenApiUtilClient.get_encode_param(node_id)
        query = {}
        if not UtilClient.is_unset(request.depth):
            query['Depth'] = request.depth
        if not UtilClient.is_unset(request.name):
            query['Name'] = request.name
        if not UtilClient.is_unset(request.order):
            query['Order'] = request.order
        if not UtilClient.is_unset(request.page_number):
            query['PageNumber'] = request.page_number
        if not UtilClient.is_unset(request.page_size):
            query['PageSize'] = request.page_size
        if not UtilClient.is_unset(request.sort_by):
            query['SortBy'] = request.sort_by
        if not UtilClient.is_unset(request.type):
            query['Type'] = request.type
        req = open_api_models.OpenApiRequest(
            headers=headers,
            query=OpenApiUtilClient.query(query)
        )
        params = open_api_models.Params(
            action='ListNodeOutputs',
            version='2021-02-02',
            protocol='HTTPS',
            pathname=f'/api/v1/runs/{run_id}/nodes/{node_id}/outputs',
            method='GET',
            auth_type='AK',
            style='ROA',
            req_body_type='json',
            body_type='json'
        )
        return TeaCore.from_map(
            paiflow_20210202_models.ListNodeOutputsResponse(),
            await self.call_api_async(params, req, runtime)
        )

    def list_node_status(
        self,
        run_id: str,
        node_id: str,
        request: paiflow_20210202_models.ListNodeStatusRequest,
    ) -> paiflow_20210202_models.ListNodeStatusResponse:
        runtime = util_models.RuntimeOptions()
        headers = {}
        return self.list_node_status_with_options(run_id, node_id, request, headers, runtime)

    async def list_node_status_async(
        self,
        run_id: str,
        node_id: str,
        request: paiflow_20210202_models.ListNodeStatusRequest,
    ) -> paiflow_20210202_models.ListNodeStatusResponse:
        runtime = util_models.RuntimeOptions()
        headers = {}
        return await self.list_node_status_with_options_async(run_id, node_id, request, headers, runtime)

    def list_node_status_with_options(
        self,
        run_id: str,
        node_id: str,
        request: paiflow_20210202_models.ListNodeStatusRequest,
        headers: Dict[str, str],
        runtime: util_models.RuntimeOptions,
    ) -> paiflow_20210202_models.ListNodeStatusResponse:
        UtilClient.validate_model(request)
        run_id = OpenApiUtilClient.get_encode_param(run_id)
        node_id = OpenApiUtilClient.get_encode_param(node_id)
        query = {}
        if not UtilClient.is_unset(request.depth):
            query['Depth'] = request.depth
        if not UtilClient.is_unset(request.type):
            query['Type'] = request.type
        req = open_api_models.OpenApiRequest(
            headers=headers,
            query=OpenApiUtilClient.query(query)
        )
        params = open_api_models.Params(
            action='ListNodeStatus',
            version='2021-02-02',
            protocol='HTTPS',
            pathname=f'/api/v1/runs/{run_id}/nodes/{node_id}/status',
            method='GET',
            auth_type='AK',
            style='ROA',
            req_body_type='json',
            body_type='json'
        )
        return TeaCore.from_map(
            paiflow_20210202_models.ListNodeStatusResponse(),
            self.call_api(params, req, runtime)
        )

    async def list_node_status_with_options_async(
        self,
        run_id: str,
        node_id: str,
        request: paiflow_20210202_models.ListNodeStatusRequest,
        headers: Dict[str, str],
        runtime: util_models.RuntimeOptions,
    ) -> paiflow_20210202_models.ListNodeStatusResponse:
        UtilClient.validate_model(request)
        run_id = OpenApiUtilClient.get_encode_param(run_id)
        node_id = OpenApiUtilClient.get_encode_param(node_id)
        query = {}
        if not UtilClient.is_unset(request.depth):
            query['Depth'] = request.depth
        if not UtilClient.is_unset(request.type):
            query['Type'] = request.type
        req = open_api_models.OpenApiRequest(
            headers=headers,
            query=OpenApiUtilClient.query(query)
        )
        params = open_api_models.Params(
            action='ListNodeStatus',
            version='2021-02-02',
            protocol='HTTPS',
            pathname=f'/api/v1/runs/{run_id}/nodes/{node_id}/status',
            method='GET',
            auth_type='AK',
            style='ROA',
            req_body_type='json',
            body_type='json'
        )
        return TeaCore.from_map(
            paiflow_20210202_models.ListNodeStatusResponse(),
            await self.call_api_async(params, req, runtime)
        )

    def list_pipeline_privileges(
        self,
        pipeline_id: str,
    ) -> paiflow_20210202_models.ListPipelinePrivilegesResponse:
        runtime = util_models.RuntimeOptions()
        headers = {}
        return self.list_pipeline_privileges_with_options(pipeline_id, headers, runtime)

    async def list_pipeline_privileges_async(
        self,
        pipeline_id: str,
    ) -> paiflow_20210202_models.ListPipelinePrivilegesResponse:
        runtime = util_models.RuntimeOptions()
        headers = {}
        return await self.list_pipeline_privileges_with_options_async(pipeline_id, headers, runtime)

    def list_pipeline_privileges_with_options(
        self,
        pipeline_id: str,
        headers: Dict[str, str],
        runtime: util_models.RuntimeOptions,
    ) -> paiflow_20210202_models.ListPipelinePrivilegesResponse:
        pipeline_id = OpenApiUtilClient.get_encode_param(pipeline_id)
        req = open_api_models.OpenApiRequest(
            headers=headers
        )
        params = open_api_models.Params(
            action='ListPipelinePrivileges',
            version='2021-02-02',
            protocol='HTTPS',
            pathname=f'/api/v1/pipelines/{pipeline_id}/privileges',
            method='GET',
            auth_type='AK',
            style='ROA',
            req_body_type='json',
            body_type='json'
        )
        return TeaCore.from_map(
            paiflow_20210202_models.ListPipelinePrivilegesResponse(),
            self.call_api(params, req, runtime)
        )

    async def list_pipeline_privileges_with_options_async(
        self,
        pipeline_id: str,
        headers: Dict[str, str],
        runtime: util_models.RuntimeOptions,
    ) -> paiflow_20210202_models.ListPipelinePrivilegesResponse:
        pipeline_id = OpenApiUtilClient.get_encode_param(pipeline_id)
        req = open_api_models.OpenApiRequest(
            headers=headers
        )
        params = open_api_models.Params(
            action='ListPipelinePrivileges',
            version='2021-02-02',
            protocol='HTTPS',
            pathname=f'/api/v1/pipelines/{pipeline_id}/privileges',
            method='GET',
            auth_type='AK',
            style='ROA',
            req_body_type='json',
            body_type='json'
        )
        return TeaCore.from_map(
            paiflow_20210202_models.ListPipelinePrivilegesResponse(),
            await self.call_api_async(params, req, runtime)
        )

    def list_pipelines(
        self,
        request: paiflow_20210202_models.ListPipelinesRequest,
    ) -> paiflow_20210202_models.ListPipelinesResponse:
        runtime = util_models.RuntimeOptions()
        headers = {}
        return self.list_pipelines_with_options(request, headers, runtime)

    async def list_pipelines_async(
        self,
        request: paiflow_20210202_models.ListPipelinesRequest,
    ) -> paiflow_20210202_models.ListPipelinesResponse:
        runtime = util_models.RuntimeOptions()
        headers = {}
        return await self.list_pipelines_with_options_async(request, headers, runtime)

    def list_pipelines_with_options(
        self,
        request: paiflow_20210202_models.ListPipelinesRequest,
        headers: Dict[str, str],
        runtime: util_models.RuntimeOptions,
    ) -> paiflow_20210202_models.ListPipelinesResponse:
        UtilClient.validate_model(request)
        query = {}
        if not UtilClient.is_unset(request.fuzzy_matching):
            query['FuzzyMatching'] = request.fuzzy_matching
        if not UtilClient.is_unset(request.page_number):
            query['PageNumber'] = request.page_number
        if not UtilClient.is_unset(request.page_size):
            query['PageSize'] = request.page_size
        if not UtilClient.is_unset(request.pipeline_identifier):
            query['PipelineIdentifier'] = request.pipeline_identifier
        if not UtilClient.is_unset(request.pipeline_provider):
            query['PipelineProvider'] = request.pipeline_provider
        if not UtilClient.is_unset(request.pipeline_version):
            query['PipelineVersion'] = request.pipeline_version
        if not UtilClient.is_unset(request.workspace_id):
            query['WorkspaceId'] = request.workspace_id
        req = open_api_models.OpenApiRequest(
            headers=headers,
            query=OpenApiUtilClient.query(query)
        )
        params = open_api_models.Params(
            action='ListPipelines',
            version='2021-02-02',
            protocol='HTTPS',
            pathname=f'/api/v1/pipelines',
            method='GET',
            auth_type='AK',
            style='ROA',
            req_body_type='json',
            body_type='json'
        )
        return TeaCore.from_map(
            paiflow_20210202_models.ListPipelinesResponse(),
            self.call_api(params, req, runtime)
        )

    async def list_pipelines_with_options_async(
        self,
        request: paiflow_20210202_models.ListPipelinesRequest,
        headers: Dict[str, str],
        runtime: util_models.RuntimeOptions,
    ) -> paiflow_20210202_models.ListPipelinesResponse:
        UtilClient.validate_model(request)
        query = {}
        if not UtilClient.is_unset(request.fuzzy_matching):
            query['FuzzyMatching'] = request.fuzzy_matching
        if not UtilClient.is_unset(request.page_number):
            query['PageNumber'] = request.page_number
        if not UtilClient.is_unset(request.page_size):
            query['PageSize'] = request.page_size
        if not UtilClient.is_unset(request.pipeline_identifier):
            query['PipelineIdentifier'] = request.pipeline_identifier
        if not UtilClient.is_unset(request.pipeline_provider):
            query['PipelineProvider'] = request.pipeline_provider
        if not UtilClient.is_unset(request.pipeline_version):
            query['PipelineVersion'] = request.pipeline_version
        if not UtilClient.is_unset(request.workspace_id):
            query['WorkspaceId'] = request.workspace_id
        req = open_api_models.OpenApiRequest(
            headers=headers,
            query=OpenApiUtilClient.query(query)
        )
        params = open_api_models.Params(
            action='ListPipelines',
            version='2021-02-02',
            protocol='HTTPS',
            pathname=f'/api/v1/pipelines',
            method='GET',
            auth_type='AK',
            style='ROA',
            req_body_type='json',
            body_type='json'
        )
        return TeaCore.from_map(
            paiflow_20210202_models.ListPipelinesResponse(),
            await self.call_api_async(params, req, runtime)
        )

    def list_runs(
        self,
        request: paiflow_20210202_models.ListRunsRequest,
    ) -> paiflow_20210202_models.ListRunsResponse:
        runtime = util_models.RuntimeOptions()
        headers = {}
        return self.list_runs_with_options(request, headers, runtime)

    async def list_runs_async(
        self,
        request: paiflow_20210202_models.ListRunsRequest,
    ) -> paiflow_20210202_models.ListRunsResponse:
        runtime = util_models.RuntimeOptions()
        headers = {}
        return await self.list_runs_with_options_async(request, headers, runtime)

    def list_runs_with_options(
        self,
        request: paiflow_20210202_models.ListRunsRequest,
        headers: Dict[str, str],
        runtime: util_models.RuntimeOptions,
    ) -> paiflow_20210202_models.ListRunsResponse:
        UtilClient.validate_model(request)
        query = {}
        if not UtilClient.is_unset(request.experiment_id):
            query['ExperimentId'] = request.experiment_id
        if not UtilClient.is_unset(request.name):
            query['Name'] = request.name
        if not UtilClient.is_unset(request.order):
            query['Order'] = request.order
        if not UtilClient.is_unset(request.page_number):
            query['PageNumber'] = request.page_number
        if not UtilClient.is_unset(request.page_size):
            query['PageSize'] = request.page_size
        if not UtilClient.is_unset(request.pipeline_id):
            query['PipelineId'] = request.pipeline_id
        if not UtilClient.is_unset(request.pipeline_ids):
            query['PipelineIds'] = request.pipeline_ids
        if not UtilClient.is_unset(request.run_id):
            query['RunId'] = request.run_id
        if not UtilClient.is_unset(request.sort_by):
            query['SortBy'] = request.sort_by
        if not UtilClient.is_unset(request.source):
            query['Source'] = request.source
        if not UtilClient.is_unset(request.status):
            query['Status'] = request.status
        if not UtilClient.is_unset(request.user_id):
            query['UserId'] = request.user_id
        if not UtilClient.is_unset(request.workspace_id):
            query['WorkspaceId'] = request.workspace_id
        req = open_api_models.OpenApiRequest(
            headers=headers,
            query=OpenApiUtilClient.query(query)
        )
        params = open_api_models.Params(
            action='ListRuns',
            version='2021-02-02',
            protocol='HTTPS',
            pathname=f'/api/v1/runs',
            method='GET',
            auth_type='AK',
            style='ROA',
            req_body_type='json',
            body_type='json'
        )
        return TeaCore.from_map(
            paiflow_20210202_models.ListRunsResponse(),
            self.call_api(params, req, runtime)
        )

    async def list_runs_with_options_async(
        self,
        request: paiflow_20210202_models.ListRunsRequest,
        headers: Dict[str, str],
        runtime: util_models.RuntimeOptions,
    ) -> paiflow_20210202_models.ListRunsResponse:
        UtilClient.validate_model(request)
        query = {}
        if not UtilClient.is_unset(request.experiment_id):
            query['ExperimentId'] = request.experiment_id
        if not UtilClient.is_unset(request.name):
            query['Name'] = request.name
        if not UtilClient.is_unset(request.order):
            query['Order'] = request.order
        if not UtilClient.is_unset(request.page_number):
            query['PageNumber'] = request.page_number
        if not UtilClient.is_unset(request.page_size):
            query['PageSize'] = request.page_size
        if not UtilClient.is_unset(request.pipeline_id):
            query['PipelineId'] = request.pipeline_id
        if not UtilClient.is_unset(request.pipeline_ids):
            query['PipelineIds'] = request.pipeline_ids
        if not UtilClient.is_unset(request.run_id):
            query['RunId'] = request.run_id
        if not UtilClient.is_unset(request.sort_by):
            query['SortBy'] = request.sort_by
        if not UtilClient.is_unset(request.source):
            query['Source'] = request.source
        if not UtilClient.is_unset(request.status):
            query['Status'] = request.status
        if not UtilClient.is_unset(request.user_id):
            query['UserId'] = request.user_id
        if not UtilClient.is_unset(request.workspace_id):
            query['WorkspaceId'] = request.workspace_id
        req = open_api_models.OpenApiRequest(
            headers=headers,
            query=OpenApiUtilClient.query(query)
        )
        params = open_api_models.Params(
            action='ListRuns',
            version='2021-02-02',
            protocol='HTTPS',
            pathname=f'/api/v1/runs',
            method='GET',
            auth_type='AK',
            style='ROA',
            req_body_type='json',
            body_type='json'
        )
        return TeaCore.from_map(
            paiflow_20210202_models.ListRunsResponse(),
            await self.call_api_async(params, req, runtime)
        )

    def list_runs_status(
        self,
        request: paiflow_20210202_models.ListRunsStatusRequest,
    ) -> paiflow_20210202_models.ListRunsStatusResponse:
        runtime = util_models.RuntimeOptions()
        headers = {}
        return self.list_runs_status_with_options(request, headers, runtime)

    async def list_runs_status_async(
        self,
        request: paiflow_20210202_models.ListRunsStatusRequest,
    ) -> paiflow_20210202_models.ListRunsStatusResponse:
        runtime = util_models.RuntimeOptions()
        headers = {}
        return await self.list_runs_status_with_options_async(request, headers, runtime)

    def list_runs_status_with_options(
        self,
        request: paiflow_20210202_models.ListRunsStatusRequest,
        headers: Dict[str, str],
        runtime: util_models.RuntimeOptions,
    ) -> paiflow_20210202_models.ListRunsStatusResponse:
        UtilClient.validate_model(request)
        body = {}
        if not UtilClient.is_unset(request.nodes):
            body['Nodes'] = request.nodes
        if not UtilClient.is_unset(request.runs):
            body['Runs'] = request.runs
        if not UtilClient.is_unset(request.workspace_id):
            body['WorkspaceId'] = request.workspace_id
        req = open_api_models.OpenApiRequest(
            headers=headers,
            body=OpenApiUtilClient.parse_to_map(body)
        )
        params = open_api_models.Params(
            action='ListRunsStatus',
            version='2021-02-02',
            protocol='HTTPS',
            pathname=f'/api/v1/runs',
            method='PUT',
            auth_type='AK',
            style='ROA',
            req_body_type='json',
            body_type='json'
        )
        return TeaCore.from_map(
            paiflow_20210202_models.ListRunsStatusResponse(),
            self.call_api(params, req, runtime)
        )

    async def list_runs_status_with_options_async(
        self,
        request: paiflow_20210202_models.ListRunsStatusRequest,
        headers: Dict[str, str],
        runtime: util_models.RuntimeOptions,
    ) -> paiflow_20210202_models.ListRunsStatusResponse:
        UtilClient.validate_model(request)
        body = {}
        if not UtilClient.is_unset(request.nodes):
            body['Nodes'] = request.nodes
        if not UtilClient.is_unset(request.runs):
            body['Runs'] = request.runs
        if not UtilClient.is_unset(request.workspace_id):
            body['WorkspaceId'] = request.workspace_id
        req = open_api_models.OpenApiRequest(
            headers=headers,
            body=OpenApiUtilClient.parse_to_map(body)
        )
        params = open_api_models.Params(
            action='ListRunsStatus',
            version='2021-02-02',
            protocol='HTTPS',
            pathname=f'/api/v1/runs',
            method='PUT',
            auth_type='AK',
            style='ROA',
            req_body_type='json',
            body_type='json'
        )
        return TeaCore.from_map(
            paiflow_20210202_models.ListRunsStatusResponse(),
            await self.call_api_async(params, req, runtime)
        )

    def start_run(
        self,
        run_id: str,
    ) -> paiflow_20210202_models.StartRunResponse:
        runtime = util_models.RuntimeOptions()
        headers = {}
        return self.start_run_with_options(run_id, headers, runtime)

    async def start_run_async(
        self,
        run_id: str,
    ) -> paiflow_20210202_models.StartRunResponse:
        runtime = util_models.RuntimeOptions()
        headers = {}
        return await self.start_run_with_options_async(run_id, headers, runtime)

    def start_run_with_options(
        self,
        run_id: str,
        headers: Dict[str, str],
        runtime: util_models.RuntimeOptions,
    ) -> paiflow_20210202_models.StartRunResponse:
        run_id = OpenApiUtilClient.get_encode_param(run_id)
        req = open_api_models.OpenApiRequest(
            headers=headers
        )
        params = open_api_models.Params(
            action='StartRun',
            version='2021-02-02',
            protocol='HTTPS',
            pathname=f'/api/v1/runs/{run_id}/start',
            method='PUT',
            auth_type='AK',
            style='ROA',
            req_body_type='json',
            body_type='json'
        )
        return TeaCore.from_map(
            paiflow_20210202_models.StartRunResponse(),
            self.call_api(params, req, runtime)
        )

    async def start_run_with_options_async(
        self,
        run_id: str,
        headers: Dict[str, str],
        runtime: util_models.RuntimeOptions,
    ) -> paiflow_20210202_models.StartRunResponse:
        run_id = OpenApiUtilClient.get_encode_param(run_id)
        req = open_api_models.OpenApiRequest(
            headers=headers
        )
        params = open_api_models.Params(
            action='StartRun',
            version='2021-02-02',
            protocol='HTTPS',
            pathname=f'/api/v1/runs/{run_id}/start',
            method='PUT',
            auth_type='AK',
            style='ROA',
            req_body_type='json',
            body_type='json'
        )
        return TeaCore.from_map(
            paiflow_20210202_models.StartRunResponse(),
            await self.call_api_async(params, req, runtime)
        )

    def terminate_run(
        self,
        run_id: str,
    ) -> paiflow_20210202_models.TerminateRunResponse:
        runtime = util_models.RuntimeOptions()
        headers = {}
        return self.terminate_run_with_options(run_id, headers, runtime)

    async def terminate_run_async(
        self,
        run_id: str,
    ) -> paiflow_20210202_models.TerminateRunResponse:
        runtime = util_models.RuntimeOptions()
        headers = {}
        return await self.terminate_run_with_options_async(run_id, headers, runtime)

    def terminate_run_with_options(
        self,
        run_id: str,
        headers: Dict[str, str],
        runtime: util_models.RuntimeOptions,
    ) -> paiflow_20210202_models.TerminateRunResponse:
        run_id = OpenApiUtilClient.get_encode_param(run_id)
        req = open_api_models.OpenApiRequest(
            headers=headers
        )
        params = open_api_models.Params(
            action='TerminateRun',
            version='2021-02-02',
            protocol='HTTPS',
            pathname=f'/api/v1/runs/{run_id}/termination',
            method='PUT',
            auth_type='AK',
            style='ROA',
            req_body_type='json',
            body_type='json'
        )
        return TeaCore.from_map(
            paiflow_20210202_models.TerminateRunResponse(),
            self.call_api(params, req, runtime)
        )

    async def terminate_run_with_options_async(
        self,
        run_id: str,
        headers: Dict[str, str],
        runtime: util_models.RuntimeOptions,
    ) -> paiflow_20210202_models.TerminateRunResponse:
        run_id = OpenApiUtilClient.get_encode_param(run_id)
        req = open_api_models.OpenApiRequest(
            headers=headers
        )
        params = open_api_models.Params(
            action='TerminateRun',
            version='2021-02-02',
            protocol='HTTPS',
            pathname=f'/api/v1/runs/{run_id}/termination',
            method='PUT',
            auth_type='AK',
            style='ROA',
            req_body_type='json',
            body_type='json'
        )
        return TeaCore.from_map(
            paiflow_20210202_models.TerminateRunResponse(),
            await self.call_api_async(params, req, runtime)
        )

    def update_pipeline(
        self,
        pipeline_id: str,
        request: paiflow_20210202_models.UpdatePipelineRequest,
    ) -> paiflow_20210202_models.UpdatePipelineResponse:
        runtime = util_models.RuntimeOptions()
        headers = {}
        return self.update_pipeline_with_options(pipeline_id, request, headers, runtime)

    async def update_pipeline_async(
        self,
        pipeline_id: str,
        request: paiflow_20210202_models.UpdatePipelineRequest,
    ) -> paiflow_20210202_models.UpdatePipelineResponse:
        runtime = util_models.RuntimeOptions()
        headers = {}
        return await self.update_pipeline_with_options_async(pipeline_id, request, headers, runtime)

    def update_pipeline_with_options(
        self,
        pipeline_id: str,
        request: paiflow_20210202_models.UpdatePipelineRequest,
        headers: Dict[str, str],
        runtime: util_models.RuntimeOptions,
    ) -> paiflow_20210202_models.UpdatePipelineResponse:
        UtilClient.validate_model(request)
        pipeline_id = OpenApiUtilClient.get_encode_param(pipeline_id)
        body = {}
        if not UtilClient.is_unset(request.manifest):
            body['Manifest'] = request.manifest
        req = open_api_models.OpenApiRequest(
            headers=headers,
            body=OpenApiUtilClient.parse_to_map(body)
        )
        params = open_api_models.Params(
            action='UpdatePipeline',
            version='2021-02-02',
            protocol='HTTPS',
            pathname=f'/api/v1/pipelines/{pipeline_id}',
            method='PUT',
            auth_type='AK',
            style='ROA',
            req_body_type='json',
            body_type='json'
        )
        return TeaCore.from_map(
            paiflow_20210202_models.UpdatePipelineResponse(),
            self.call_api(params, req, runtime)
        )

    async def update_pipeline_with_options_async(
        self,
        pipeline_id: str,
        request: paiflow_20210202_models.UpdatePipelineRequest,
        headers: Dict[str, str],
        runtime: util_models.RuntimeOptions,
    ) -> paiflow_20210202_models.UpdatePipelineResponse:
        UtilClient.validate_model(request)
        pipeline_id = OpenApiUtilClient.get_encode_param(pipeline_id)
        body = {}
        if not UtilClient.is_unset(request.manifest):
            body['Manifest'] = request.manifest
        req = open_api_models.OpenApiRequest(
            headers=headers,
            body=OpenApiUtilClient.parse_to_map(body)
        )
        params = open_api_models.Params(
            action='UpdatePipeline',
            version='2021-02-02',
            protocol='HTTPS',
            pathname=f'/api/v1/pipelines/{pipeline_id}',
            method='PUT',
            auth_type='AK',
            style='ROA',
            req_body_type='json',
            body_type='json'
        )
        return TeaCore.from_map(
            paiflow_20210202_models.UpdatePipelineResponse(),
            await self.call_api_async(params, req, runtime)
        )

    def update_pipeline_privileges(
        self,
        pipeline_id: str,
        request: paiflow_20210202_models.UpdatePipelinePrivilegesRequest,
    ) -> paiflow_20210202_models.UpdatePipelinePrivilegesResponse:
        runtime = util_models.RuntimeOptions()
        headers = {}
        return self.update_pipeline_privileges_with_options(pipeline_id, request, headers, runtime)

    async def update_pipeline_privileges_async(
        self,
        pipeline_id: str,
        request: paiflow_20210202_models.UpdatePipelinePrivilegesRequest,
    ) -> paiflow_20210202_models.UpdatePipelinePrivilegesResponse:
        runtime = util_models.RuntimeOptions()
        headers = {}
        return await self.update_pipeline_privileges_with_options_async(pipeline_id, request, headers, runtime)

    def update_pipeline_privileges_with_options(
        self,
        pipeline_id: str,
        request: paiflow_20210202_models.UpdatePipelinePrivilegesRequest,
        headers: Dict[str, str],
        runtime: util_models.RuntimeOptions,
    ) -> paiflow_20210202_models.UpdatePipelinePrivilegesResponse:
        UtilClient.validate_model(request)
        pipeline_id = OpenApiUtilClient.get_encode_param(pipeline_id)
        body = {}
        if not UtilClient.is_unset(request.users):
            body['Users'] = request.users
        req = open_api_models.OpenApiRequest(
            headers=headers,
            body=OpenApiUtilClient.parse_to_map(body)
        )
        params = open_api_models.Params(
            action='UpdatePipelinePrivileges',
            version='2021-02-02',
            protocol='HTTPS',
            pathname=f'/api/v1/pipelines/{pipeline_id}/privileges',
            method='PUT',
            auth_type='AK',
            style='ROA',
            req_body_type='json',
            body_type='json'
        )
        return TeaCore.from_map(
            paiflow_20210202_models.UpdatePipelinePrivilegesResponse(),
            self.call_api(params, req, runtime)
        )

    async def update_pipeline_privileges_with_options_async(
        self,
        pipeline_id: str,
        request: paiflow_20210202_models.UpdatePipelinePrivilegesRequest,
        headers: Dict[str, str],
        runtime: util_models.RuntimeOptions,
    ) -> paiflow_20210202_models.UpdatePipelinePrivilegesResponse:
        UtilClient.validate_model(request)
        pipeline_id = OpenApiUtilClient.get_encode_param(pipeline_id)
        body = {}
        if not UtilClient.is_unset(request.users):
            body['Users'] = request.users
        req = open_api_models.OpenApiRequest(
            headers=headers,
            body=OpenApiUtilClient.parse_to_map(body)
        )
        params = open_api_models.Params(
            action='UpdatePipelinePrivileges',
            version='2021-02-02',
            protocol='HTTPS',
            pathname=f'/api/v1/pipelines/{pipeline_id}/privileges',
            method='PUT',
            auth_type='AK',
            style='ROA',
            req_body_type='json',
            body_type='json'
        )
        return TeaCore.from_map(
            paiflow_20210202_models.UpdatePipelinePrivilegesResponse(),
            await self.call_api_async(params, req, runtime)
        )

    def update_run(
        self,
        run_id: str,
        request: paiflow_20210202_models.UpdateRunRequest,
    ) -> paiflow_20210202_models.UpdateRunResponse:
        runtime = util_models.RuntimeOptions()
        headers = {}
        return self.update_run_with_options(run_id, request, headers, runtime)

    async def update_run_async(
        self,
        run_id: str,
        request: paiflow_20210202_models.UpdateRunRequest,
    ) -> paiflow_20210202_models.UpdateRunResponse:
        runtime = util_models.RuntimeOptions()
        headers = {}
        return await self.update_run_with_options_async(run_id, request, headers, runtime)

    def update_run_with_options(
        self,
        run_id: str,
        request: paiflow_20210202_models.UpdateRunRequest,
        headers: Dict[str, str],
        runtime: util_models.RuntimeOptions,
    ) -> paiflow_20210202_models.UpdateRunResponse:
        UtilClient.validate_model(request)
        run_id = OpenApiUtilClient.get_encode_param(run_id)
        body = {}
        if not UtilClient.is_unset(request.name):
            body['Name'] = request.name
        req = open_api_models.OpenApiRequest(
            headers=headers,
            body=OpenApiUtilClient.parse_to_map(body)
        )
        params = open_api_models.Params(
            action='UpdateRun',
            version='2021-02-02',
            protocol='HTTPS',
            pathname=f'/api/v1/runs/{run_id}',
            method='PUT',
            auth_type='AK',
            style='ROA',
            req_body_type='json',
            body_type='json'
        )
        return TeaCore.from_map(
            paiflow_20210202_models.UpdateRunResponse(),
            self.call_api(params, req, runtime)
        )

    async def update_run_with_options_async(
        self,
        run_id: str,
        request: paiflow_20210202_models.UpdateRunRequest,
        headers: Dict[str, str],
        runtime: util_models.RuntimeOptions,
    ) -> paiflow_20210202_models.UpdateRunResponse:
        UtilClient.validate_model(request)
        run_id = OpenApiUtilClient.get_encode_param(run_id)
        body = {}
        if not UtilClient.is_unset(request.name):
            body['Name'] = request.name
        req = open_api_models.OpenApiRequest(
            headers=headers,
            body=OpenApiUtilClient.parse_to_map(body)
        )
        params = open_api_models.Params(
            action='UpdateRun',
            version='2021-02-02',
            protocol='HTTPS',
            pathname=f'/api/v1/runs/{run_id}',
            method='PUT',
            auth_type='AK',
            style='ROA',
            req_body_type='json',
            body_type='json'
        )
        return TeaCore.from_map(
            paiflow_20210202_models.UpdateRunResponse(),
            await self.call_api_async(params, req, runtime)
        )



#  Copyright 2023 Alibaba, Inc. or its affiliates.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import absolute_import

import functools
import importlib.util
import random
import re
import socket
import string
import sys
import time
import warnings
from datetime import datetime
from functools import lru_cache
from typing import Callable, Dict, List, Optional, Union

from semantic_version import Version

from pai.common.consts import (
    INSTANCE_TYPE_LOCAL,
    INSTANCE_TYPE_LOCAL_GPU,
    FileSystemInputScheme,
)
from pai.version import VERSION

DEFAULT_PLAIN_TEXT_ALLOW_CHARACTERS = string.ascii_letters + string.digits + "_"


def is_iterable(arg):
    try:
        _ = iter(arg)
        return True
    except TypeError:
        return False


def random_str(n):
    """Random string generation with lower case letters and digits.

    Args:
        n: Size of generated random string.

    Returns:
        str: generated random string.

    """
    return "".join(
        random.choice(string.ascii_lowercase + string.digits) for _ in range(n)
    )


def camel_to_snake(name):
    """Convert a name from camel case to snake case."""
    name = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", name)
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", name).lower()


def snake_to_camel(name):
    """Convert a name from snake case to camel case."""
    return "".join([w.title() for w in name.split("_")])


def make_list_resource_iterator(method: Callable, **kwargs):
    """Wrap resource list method as an iterator.

    Args:
        method: Resource List method.
        **kwargs: arguments for the method.

    Yields:
        A resource iterator.
    """

    from pai.api.base import PaginatedResult

    page_number = kwargs.get("page_number", 1)
    page_size = kwargs.get("page_size", 10)

    while True:
        kwargs.update(page_number=page_number, page_size=page_size)
        result = method(**kwargs)
        if isinstance(result, PaginatedResult):
            total_count = result.total_count
            result = result.items
        else:
            total_count = None
        for item in result:
            yield item

        if len(result) == 0 or len(result) < page_size:
            return
        if total_count and page_number * page_size >= total_count:
            return

        page_number += 1


def to_plain_text(
    input_str: str, allowed_characters=DEFAULT_PLAIN_TEXT_ALLOW_CHARACTERS, repl_ch="_"
):
    """Replace characters in input_str if it is not in allowed_characters."""
    return "".join([c if c in allowed_characters else repl_ch for c in input_str])


def http_user_agent(user_agent: Optional[Union[Dict, str]] = None) -> str:
    """Generate HTTP User-Agent that represents current client."""
    ua = f"pai-python-sdk/{VERSION}; python/{sys.version.split()[0]}"
    if isinstance(user_agent, dict):
        ua += "; " + "; ".join(f"{k}/{v}" for k, v in user_agent.items())
    elif isinstance(user_agent, str):
        ua += "; " + user_agent
    return ua


def is_notebook() -> bool:
    """Return True if current environment is notebook."""
    try:
        from IPython import get_ipython

        shell = get_ipython().__class__.__name__
        for parent_cls in shell.__mro__:
            if parent_cls.__name__ == "ZMQInteractiveShell":
                return True
        return False
    except (NameError, ImportError):
        return False


def is_local_run_instance_type(instance_type: str) -> bool:
    """Return True if instance_type is local run instance type."""
    return instance_type and instance_type.strip() in [
        INSTANCE_TYPE_LOCAL_GPU,
        INSTANCE_TYPE_LOCAL,
    ]


def generate_repr(repr_obj, *attr_names: str, **kwargs) -> str:
    """Generate a string representation of the given object.

    Args:
        repr_obj: The object used to generate the string representation.
        attr_names: A list of attribute names to include in the string representation.

    Returns:
        str: A string representation of the object.

    """
    attrs = {name: getattr(repr_obj, name) for name in attr_names}
    attrs.update(kwargs)
    attr_repr = ", ".join(["{}={}".format(k, v) for k, v in attrs.items()])
    cls_name = repr_obj.__class__.__name__

    return f"{cls_name}({attr_repr})"


def to_semantic_version(version_str: str) -> Version:
    """Convert version_str to semantic version.

    Convert version_str to semantic version, if version_str is not a valid
    semantic version, return '0.0.0'.

    Args:
        version_str[str]: Version string, such as '1.0.0', '1.0.0-rc1', '1.0.0+build.1'.

    Returns:
        :class:`semantic_version.Version`: Semantic version.
    """
    try:
        return Version.coerce(version_str)
    except ValueError:
        # if version_str is not a valid semantic version, return '0.0.0'
        return Version.coerce("0.0.0")


def is_odps_table_uri(uri: str) -> bool:
    """Return True if uri is an odps table input URI.

    Args:
        uri (str): URI of input table, such as 'odps://<project_name>/tables/<table_name>'.

    Examples:
        >>> is_odps_table_uri('odps://<project_name>/tables/<table_name>')
        True

    """
    if not uri.startswith("odps://"):
        return False
    info = uri[7:].split("/", 2)
    if len(info) != 3:
        return False
    return info[1] == "tables"


def is_filesystem_uri(uri: str) -> bool:
    """Return True if uri is a filesystem input URI.

    Args:
        uri (str): URI of input NAS, such as 'nas://<FileSystemId>/path/to/data/directory/'.

    Examples:
        # Standard or Extreme file system type
        >>> is_filesystem_uri('nas://<FileSystemId>/path/to/data/directory/')
        True
        # CPFS file system type
        >>> is_filesystem_uri('cpfs://<FileSystemId>/<ProtocolServiceId>/<ExportId>')
        True

    """
    schemas = {
        v for k, v in FileSystemInputScheme.__dict__.items() if not k.startswith("__")
    }

    return any(uri.startswith(f"{schema}://") for schema in schemas)


def is_dataset_id(item: str) -> bool:
    """Return True if given input is a dataset ID.

    Args:
        item (str): user input dataset ID or Dataset ID and dataset version, separated by a slash.

    Examples:
        >>> is_dataset_id('d-ybko3rap60c4gs9flc')
        True
        >>> is_dataset_id('d-ybko3rap60c4gs9flc/v1')
        True
    """
    return item.startswith("d-")


def is_nas_uri(uri: Union[str, bytes]) -> bool:
    """Determines whether the given uri is a NAS uri.

    Args:
        uri (Union[str, bytes]): A string in NAS URI schema:
            nas://29**d-b12****446.cn-hangzhou.nas.aliyuncs.com/data/path/
            nas://29****123-y**r.cn-hangzhou.extreme.nas.aliyuncs.com/data/path/


    Returns:
        bool: True if the given uri is a NAS uri, else False.

    """
    return bool(uri and isinstance(uri, (str, bytes)) and str(uri).startswith("nas://"))


@lru_cache()
def is_domain_connectable(domain: str, port: int = 80, timeout: int = 1) -> bool:
    """Check if the domain is connectable."""

    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    # Set the timeout for the socket
    sock.settimeout(timeout)
    try:
        # Get the IP address of the domain
        ip = socket.gethostbyname(domain)
        # Try to connect to the IP address on specific port (default 80, HTTP)
        sock.connect((ip, port))
        # If the connection is successful, return True
        return True
    except (socket.timeout, socket.gaierror, socket.error):
        # If there is an error connecting, return False
        return False
    finally:
        # Close the socket
        sock.close()


def experimental(callable_entity):
    """Decorator to mark functions or classes as experimental"""

    @functools.wraps(callable_entity)
    def wrapper(*args, **kwargs):
        message = f"{callable_entity.__name__} is experimental and may change or be removed in future releases."
        warnings.warn(message, category=FutureWarning, stacklevel=2)
        return callable_entity(*args, **kwargs)

    return wrapper


def retry(max_attempts=3, wait_secs=1, exceptions=(Exception,), report_retries=True):
    """Decorator to make functions retry by config"""

    def decorator_retry(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            attempts = 0
            while attempts < max_attempts:
                try:
                    result = func(*args, **kwargs)
                    return result
                except exceptions as e:
                    attempts += 1
                    if attempts == max_attempts:
                        raise  # Re-raise the last exception when all the attempts have failed
                    if report_retries:
                        warnings.warn(f"Retry {attempts}/{max_attempts} failed: {e}")
                    time.sleep(wait_secs)

        return wrapper

    return decorator_retry


def print_table(headers: List[str], rows: List[List[str]]):
    """Give headers and rows, print as table to stdout."""

    length = len(headers)
    for row in rows:
        if len(row) != length:
            raise ValueError("Unable to print table, headers length mismatch with rows")

    column_widths = [
        max(len(str(value)) for value in column) for column in zip(headers, *rows)
    ]
    header_row = " | ".join(
        f"{header:<{column_widths[i]}}" for i, header in enumerate(headers)
    )

    print(header_row)
    print("-" * len(header_row))
    for row in rows:
        print(
            " | ".join(
                f"{str(value):<{column_widths[i]}}" for i, value in enumerate(row)
            )
        )


def is_package_available(package_name: str) -> bool:
    """Check if the package is available in the current environment."""
    return True if importlib.util.find_spec(package_name) is not None else False


def timestamp(sep: str = "-", utc: bool = False) -> str:
    """Return a timestamp with millisecond precision.

    Args:
        sep: The separator between date and time.
        utc: Whether to use UTC time.

    Returns:
        str: A timestamp with millisecond precision.

    """
    if utc:
        res = datetime.utcnow().strftime("%Y%m%d-%H%M%S-%f")[:-3]
    else:
        res = datetime.now().strftime("%Y%m%d-%H%M%S-%f")[:-3]
    if sep != "-":
        res = res.replace("-", sep)
    return res


def name_from_base(base_name: str, sep: str = "-") -> str:
    """Return a name with base_name and timestamp.

    Args:
        base_name: The base name of the returned name.
        sep: The separator between base_name and timestamp.

    Returns:
        str: A name with base_name and timestamp.

    """
    return "{base_name}{sep}{timestamp}".format(
        base_name=base_name, sep=sep, timestamp=timestamp(sep=sep, utc=False)
    )



#  Copyright 2023 Alibaba, Inc. or its affiliates.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Any, Dict, List, Optional

from ..api.base import PaginatedResult, ServiceName, WorkspaceScopedResourceAPI
from ..libs.alibabacloud_paistudio20220112.models import (
    AlgorithmSpec,
    CreateTrainingJobRequest,
    CreateTrainingJobRequestComputeResource,
    CreateTrainingJobRequestComputeResourceInstanceSpec,
    CreateTrainingJobRequestComputeResourceSpotSpec,
    CreateTrainingJobRequestExperimentConfig,
    CreateTrainingJobRequestHyperParameters,
    CreateTrainingJobRequestInputChannels,
    CreateTrainingJobRequestLabels,
    CreateTrainingJobRequestOutputChannels,
    CreateTrainingJobRequestScheduler,
    CreateTrainingJobRequestSettings,
    CreateTrainingJobRequestUserVpc,
    CreateTrainingJobResponseBody,
    GetTrainingJobRequest,
    GetTrainingJobResponseBody,
    ListTrainingJobLogsRequest,
    ListTrainingJobLogsResponseBody,
    ListTrainingJobsRequest,
)


class TrainingJobAPI(WorkspaceScopedResourceAPI):
    BACKEND_SERVICE_NAME = ServiceName.PAI_STUDIO

    _list_method = "list_training_jobs_with_options"
    _create_method = "create_training_job_with_options"
    _get_method = "get_training_job_with_options"
    _list_logs_method = "list_training_job_logs_with_options"

    # _list_method = "list_training_jobs_with_options"

    def list(
        self,
        page_size: int = 20,
        page_number: int = 1,
        order: str = None,
        sort_by: str = None,
        status: str = None,
        training_job_name: str = None,
    ) -> PaginatedResult:
        request = ListTrainingJobsRequest(
            page_size=page_size,
            page_number=page_number,
            status=status,
            training_job_name=training_job_name,
            order=order,
            sort_by=sort_by,
        )
        res = self._do_request(
            method_=self._list_method,
            tmp_req=request,
        )

        return self.make_paginated_result(res)

    def get_api_object_by_resource_id(self, resource_id) -> Dict[str, Any]:
        res: GetTrainingJobResponseBody = self._do_request(
            method_=self._get_method,
            training_job_id=resource_id,
            request=GetTrainingJobRequest(),
        )
        return res.to_map()

    def get(self, training_job_id) -> Dict[str, Any]:
        return self.get_api_object_by_resource_id(training_job_id)

    def create(
        self,
        instance_type,
        instance_count,
        job_name,
        spot_spec: Optional[Dict[str, Any]] = None,
        instance_spec: Optional[Dict[str, str]] = None,
        resource_id: Optional[str] = None,
        resource_type: Optional[str] = None,
        hyperparameters: Optional[Dict[str, Any]] = None,
        input_channels: Optional[List[Dict[str, Any]]] = None,
        output_channels: Optional[List[Dict[str, Any]]] = None,
        environments: Dict[str, str] = None,
        requirements: List[str] = None,
        labels: Optional[Dict[str, str]] = None,
        max_running_in_seconds: Optional[int] = None,
        description: Optional[str] = None,
        algorithm_name: Optional[str] = None,
        algorithm_version: Optional[str] = None,
        algorithm_provider: Optional[str] = None,
        algorithm_spec: Optional[Dict[str, Any]] = None,
        user_vpc_config: Optional[Dict[str, Any]] = None,
        experiment_config: Optional[Dict[str, Any]] = None,
        settings: Optional[Dict[str, Any]] = None,
    ) -> str:
        """Create a TrainingJob."""
        if algorithm_spec and (
            algorithm_name or algorithm_version or algorithm_provider
        ):
            raise ValueError(
                "Please provide algorithm_spec or a tuple of (algorithm_name, "
                "algorithm_version or algorithm_provider), but not both."
            )

        if algorithm_spec:
            algo_spec = AlgorithmSpec().from_map(algorithm_spec)
        else:
            algo_spec = None

        input_channels = [
            CreateTrainingJobRequestInputChannels().from_map(ch)
            for ch in input_channels
        ]
        output_channels = [
            CreateTrainingJobRequestOutputChannels().from_map(ch)
            for ch in output_channels
        ]
        if instance_type:
            spot_spec = (
                CreateTrainingJobRequestComputeResourceSpotSpec().from_map(spot_spec)
                if spot_spec
                else None
            )
            compute_resource = CreateTrainingJobRequestComputeResource(
                ecs_count=instance_count,
                ecs_spec=instance_type,
                # use_spot_instance=bool(spot_spec),
                spot_spec=spot_spec,
            )
        elif instance_spec:
            compute_resource = CreateTrainingJobRequestComputeResource(
                resource_id=resource_id,
                instance_count=instance_count,
                instance_spec=CreateTrainingJobRequestComputeResourceInstanceSpec().from_map(
                    instance_spec
                ),
            )
        else:
            raise ValueError("Please provide instance_type or instance_spec.")

        hyperparameters = hyperparameters or dict()
        hyper_parameters = [
            CreateTrainingJobRequestHyperParameters(
                name=name,
                value=str(value),
            )
            for name, value in hyperparameters.items()
        ]
        labels = (
            [
                CreateTrainingJobRequestLabels(key=key, value=value)
                for key, value in labels.items()
            ]
            if labels
            else None
        )

        scheduler = CreateTrainingJobRequestScheduler(
            max_running_time_in_seconds=max_running_in_seconds
        )

        request = CreateTrainingJobRequest(
            algorithm_name=algorithm_name,
            algorithm_provider=algorithm_provider,
            algorithm_version=algorithm_version,
            compute_resource=compute_resource,
            hyper_parameters=hyper_parameters,
            input_channels=input_channels,
            resource_type=resource_type,
            environments=environments,
            python_requirements=requirements,
            labels=labels,
            output_channels=output_channels,
            scheduler=scheduler,
            training_job_description=description,
            training_job_name=job_name,
            algorithm_spec=algo_spec,
            user_vpc=CreateTrainingJobRequestUserVpc().from_map(user_vpc_config),
            experiment_config=CreateTrainingJobRequestExperimentConfig().from_map(
                experiment_config
            ),
            settings=(
                CreateTrainingJobRequestSettings().from_map(settings)
                if settings
                else None
            ),
        )

        resp: CreateTrainingJobResponseBody = self._do_request(
            method_=self._create_method, request=request
        )

        return resp.training_job_id

    def list_logs(
        self,
        training_job_id,
        worker_id=None,
        page_size=10,
        page_number=1,
        start_time=None,
        end_time=None,
    ) -> PaginatedResult:
        request = ListTrainingJobLogsRequest(
            page_size=page_size,
            page_number=page_number,
            start_time=start_time,
            end_time=end_time,
            worker_id=worker_id,
        )
        resp: ListTrainingJobLogsResponseBody = self._do_request(
            method_=self._list_logs_method,
            training_job_id=training_job_id,
            request=request,
        )
        # resp.logs may be None
        logs = resp.logs or []
        total_count = resp.total_count or 0
        return PaginatedResult(
            items=logs,
            total_count=total_count,
        )



#  Copyright 2024 Alibaba, Inc. or its affiliates.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Any, Dict, List, Optional

from .common import ProviderAlibabaPAI
from .common.logging import get_logger
from .common.utils import make_list_resource_iterator
from .session import Session, get_default_session

logger = get_logger(__name__)


def list_common_datasets(
    name: str = None,
    session: Optional[Session] = None,
) -> List[Dict[str, Any]]:
    session = session or get_default_session()

    gen = make_list_resource_iterator(
        session.dataset_api.list,
        name=name,
        provider=ProviderAlibabaPAI,
        # set the workspace_id manually, prevent using the default workspace of the
        # session.
        workspace_id=0,
        order="DESC",
    )

    return gen



#  Copyright 2023 Alibaba, Inc. or its affiliates.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import copy
from abc import ABCMeta, abstractmethod
from collections import Counter

import six

from .artifact import LocationArtifactMetadata, PipelineArtifact
from .parameter import PipelineParameter

IO_TYPE_INPUTS = "inputs"
IO_TYPE_OUTPUTS = "outputs"


def sort_variable_by_category(items):
    """Sort variables by category."""
    if not items:
        return [], [], []
    counter = Counter((item.name for item in items))
    conflicts = {key for key, count in counter.items() if count > 1}
    if conflicts:
        raise ValueError("Parameter/Artifact names conflict:%s" % (",".join(conflicts)))

    arts = [item for item in items if item.variable_category == "artifacts"]
    params = [item for item in items if item.variable_category == "parameters"]
    return params, arts, params + arts


class IndexedItemMixin(six.with_metaclass(ABCMeta, object)):
    def __init__(self, items):
        self._items = items
        self._indexer = {self.index_key(item): idx for idx, item in enumerate(items)}

    def __getitem__(self, key):
        if isinstance(key, six.integer_types):
            return self._items[key]
        elif isinstance(key, six.string_types):
            return self._items[self._indexer[key]]
        elif isinstance(key, slice):
            return self._items.__getitem__(key)

    def __iter__(self):
        return iter(self._items)

    def __len__(self):
        return len(self._items)

    @abstractmethod
    def index_key(self, item):
        pass


class IOSpecBase(IndexedItemMixin):
    """Inputs/Outputs spec base."""

    def __init__(self, items):
        parameter_items, artifact_items, items = sort_variable_by_category(items)
        self._parameters = Parameters(parameter_items)
        self._artifacts = Artifacts(artifact_items)
        super(IOSpecBase, self).__init__(items)

    @staticmethod
    def sort_items(items):
        # ensure parameters is prior to artifacts
        # `sorted` in Python is stable sort.
        return sorted(items, key=lambda x: 0 if isinstance(x, PipelineParameter) else 1)

    @property
    def items(self):
        return self._items

    @property
    def artifacts(self):
        return self._artifacts

    @property
    def parameters(self):
        return self._parameters

    def __repr__(self):
        return "%s:\n%s" % (
            type(self).__name__,
            "\n".join(["\t" + str(item) for item in self._items]),
        )

    def to_dict(self):
        af_pos = next(
            (
                idx
                for idx, item in enumerate(self._items)
                if item.variable_category == "artifacts"
            ),
            len(self._items),
        )

        d = {
            "parameters": [param.to_dict() for param in self._items[:af_pos]],
            "artifacts": [af.to_dict() for af in self._items[af_pos:]],
        }
        return d

    def index_key(self, item):
        return item.name


class InputsSpec(IOSpecBase):
    """Inputs spec for"""

    def __init__(self, inputs):
        super(InputsSpec, self).__init__(items=inputs)

    def assign(self, inputs_args):
        """

        Args:
            inputs_args:
        """
        assign_items = []
        if isinstance(inputs_args, list):
            for idx, arg in enumerate(inputs_args):
                self._items[idx].assign(arg)
                assign_items.append(self.items[idx])
        elif isinstance(inputs_args, dict):
            for k, v in inputs_args.items():
                self._items[self._indexer[k]].assign(v)
                assign_items.append(self._items[self._indexer[k]])
        else:
            raise ValueError(
                "Unexpected input_args type:%s, required list or dict"
                % type(inputs_args)
            )
        return assign_items


class OutputsSpec(IOSpecBase):
    def __init__(self, outputs):
        super(OutputsSpec, self).__init__(items=outputs)
        for item in self.items:
            item.kind = IO_TYPE_OUTPUTS


class Parameters(IndexedItemMixin):
    def index_key(self, item):
        return item.name


class Artifacts(IndexedItemMixin):
    def index_key(self, item):
        return item.name


def load_input_output_spec(p, spec):
    inputs = []
    outputs = []
    spec = copy.deepcopy(spec)
    for param in spec["inputs"].get("parameters", []):
        inputs.append(_load_parameter_spec(p, param.copy(), "inputs"))

    for af in spec["inputs"].get("artifacts", []):
        inputs.append(_load_artifact_spec(p, af, "inputs"))

    for param in spec["outputs"].get("parameters", []):
        outputs.append(_load_parameter_spec(p, param, "outputs"))

    for af in spec["outputs"].get("artifacts", []):
        outputs.append(_load_artifact_spec(p, af, "outputs"))
    return InputsSpec(inputs), OutputsSpec(outputs)


def _load_parameter_spec(p, param_spec, io_type):
    typ = param_spec.pop("type", None)
    name = param_spec.pop("name")
    from_ = param_spec.pop("from", None)
    feasible = param_spec.pop("feasible", None)
    value = param_spec.pop("value", None)
    desc = param_spec.pop("desc", None)

    param = PipelineParameter(
        name=name,
        typ=typ,
        default=value,
        desc=desc,
        io_type=io_type,
        from_=from_,
        parent=p,
        feasible=feasible,
    )
    return param


def _load_artifact_spec(p, artifact_spec, io_type):
    assert io_type in ("inputs", "outputs")
    metadata = LocationArtifactMetadata.from_dict(artifact_spec.get("metadata", None))

    name = artifact_spec.get("name", None)
    from_ = artifact_spec.get("from", None)
    value = artifact_spec.get("value", None)
    desc = artifact_spec.get("desc", None)
    required = artifact_spec.get("required", False)
    repeated = artifact_spec.get("repeated", False)

    af = PipelineArtifact(
        name=name,
        metadata=metadata,
        io_type=io_type,
        parent=p,
        from_=from_,
        value=value,
        desc=desc,
        required=required,
        repeated=repeated,
    )
    return af



#  Copyright 2023 Alibaba, Inc. or its affiliates.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import typing
from typing import Any, Dict

from ..libs.alibabacloud_aiworkspace20210204.models import (
    CreateModelRequest,
    CreateModelVersionRequest,
    Label,
    ListModelsRequest,
    ListModelsResponseBody,
    ListModelVersionsRequest,
    ListModelVersionsResponseBody,
    UpdateModelVersionRequest,
)
from .base import PaginatedResult, ServiceName, WorkspaceScopedResourceAPI

if typing.TYPE_CHECKING:
    pass


class ModelAPI(WorkspaceScopedResourceAPI):

    BACKEND_SERVICE_NAME = ServiceName.PAI_WORKSPACE

    _create_model_method = "create_model_with_options"
    _list_model_method = "list_models_with_options"
    _get_model_method = "get_model_with_options"
    _delete_model_method = "delete_model_with_options"

    _create_model_version_method = "create_model_version_with_options"
    _list_model_version_method = "list_model_versions_with_options"
    _get_model_version_method = "get_model_version_with_options"
    _update_model_version_method = "update_model_version_with_options"
    _delete_model_version_method = "delete_model_version_with_options"

    def create(
        self,
        accessibility: str = None,
        domain: str = None,
        labels: Dict[str, str] = None,
        model_description: str = None,
        model_doc: str = None,
        model_name: str = None,
        origin: str = None,
        task: str = None,
        workspace_id: str = None,
    ) -> str:
        labels = [Label(key=k, value=v) for k, v in labels.items()] if labels else []

        request = CreateModelRequest(
            accessibility=accessibility,
            domain=domain,
            labels=labels,
            model_description=model_description,
            model_doc=model_doc,
            model_name=model_name,
            origin=origin,
            task=task,
            workspace_id=workspace_id,
        )

        resp = self._do_request(self._create_model_method, request=request)
        return resp.model_id

    def list(
        self,
        collections: str = None,
        domain: str = None,
        label: str = None,
        label_string: str = None,
        labels: str = None,
        model_name: str = None,
        order: str = None,
        origin: str = None,
        page_number: int = None,
        page_size: int = None,
        provider: str = None,
        query: str = None,
        sort_by: str = None,
        task: str = None,
        workspace_id: str = None,
    ) -> PaginatedResult:
        request = ListModelsRequest(
            collections=collections,
            domain=domain,
            label=label,
            label_string=label_string,
            labels=labels,
            model_name=model_name,
            order=order,
            origin=origin,
            page_number=page_number,
            page_size=page_size,
            provider=provider,
            query=query,
            sort_by=sort_by,
            task=task,
            workspace_id=workspace_id,
        )

        resp: ListModelsResponseBody = self._do_request(
            self._list_model_method, request=request
        )
        return self.make_paginated_result(resp)

    def get(self, model_id: str):
        resp = self._do_request(method_=self._get_model_method, model_id=model_id)
        return resp.to_map()

    def delete(self, model_id: str):
        self._do_request(method_=self._delete_model_method, model_id=model_id)

    def create_version(
        self,
        model_id: str,
        approval_status: str = None,
        evaluation_spec: Dict[str, Any] = None,
        format_type: str = None,
        framework_type: str = None,
        inference_spec: Dict[str, Any] = None,
        labels: Dict[str, str] = None,
        metrics: Dict[str, Any] = None,
        options: str = None,
        source_id: str = None,
        source_type: str = None,
        training_spec: Dict[str, Any] = None,
        uri: str = None,
        version_description: str = None,
        version_name: str = None,
    ):
        """Create a ModeVersion resource."""
        labels = [Label(key=k, value=v) for k, v in labels.items()] if labels else []

        request = CreateModelVersionRequest(
            approval_status=approval_status,
            evaluation_spec=evaluation_spec,
            format_type=format_type,
            framework_type=framework_type,
            inference_spec=inference_spec,
            labels=labels,
            metrics=metrics,
            options=options,
            source_id=source_id,
            source_type=source_type,
            training_spec=training_spec,
            uri=uri,
            version_description=version_description,
            version_name=version_name,
        )

        response = self._do_request(
            self._create_model_version_method, model_id=model_id, request=request
        )

        version_name = response.to_map()["VersionName"]
        return version_name

    def list_versions(
        self,
        model_id,
        approval_status: str = None,
        format_type: str = None,
        framework_type: str = None,
        label: str = None,
        label_string: str = None,
        labels: str = None,
        order: str = None,
        page_number: int = None,
        page_size: int = None,
        sort_by: str = None,
        source_id: str = None,
        source_type: str = None,
        version_name: str = None,
    ) -> PaginatedResult:
        request = ListModelVersionsRequest(
            approval_status=approval_status,
            format_type=format_type,
            framework_type=framework_type,
            label=label,
            label_string=label_string,
            labels=labels,
            order=order,
            page_number=page_number,
            page_size=page_size,
            sort_by=sort_by,
            source_id=source_id,
            source_type=source_type,
            version_name=version_name,
        )

        resp: ListModelVersionsResponseBody = self._do_request(
            self._list_model_version_method, model_id=model_id, request=request
        )

        data = resp.to_map()
        for v in data["Versions"]:
            v.update(
                {
                    "ModelId": model_id,
                }
            )
        return self.make_paginated_result(data)

    def get_version(self, model_id: str, version: str):
        resp = self._do_request(
            self._get_model_version_method, model_id=model_id, version_name=version
        )
        obj = resp.to_map()
        obj.update({"ModelId": model_id})
        return obj

    def update_version(
        self,
        model_id: str,
        version: str,
        approval_status: str = None,
        evaluation_spec: Dict[str, Any] = None,
        inference_spec: Dict[str, Any] = None,
        metrics: Dict[str, Any] = None,
        options: str = None,
        source_id: str = None,
        source_type: str = None,
        training_spec: Dict[str, Any] = None,
        version_description: str = None,
    ):
        request = UpdateModelVersionRequest(
            approval_status=approval_status,
            evaluation_spec=evaluation_spec,
            inference_spec=inference_spec,
            metrics=metrics,
            options=options,
            source_id=source_id,
            source_type=source_type,
            training_spec=training_spec,
            version_description=version_description,
        )
        self._do_request(
            self._update_model_version_method,
            model_id=model_id,
            version_name=version,
            request=request,
        )

    def delete_version(self, model_id: str, version: str):
        self._do_request(
            self._delete_model_version_method,
            model_id=model_id,
            version_name=version,
        )



#  Copyright 2023 Alibaba, Inc. or its affiliates.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import json
from typing import Any, Dict, List, Optional, Union

from ..common.consts import (
    DatasetSourceType,
    DataType,
    FileProperty,
    ResourceAccessibility,
)
from ..libs.alibabacloud_aiworkspace20210204.models import (
    CreateDatasetRequest,
    DatasetLabel,
    ListDatasetsRequest,
    ListDatasetsResponseBody,
)
from .base import PaginatedResult, ServiceName, WorkspaceScopedResourceAPI


class DatasetAPI(WorkspaceScopedResourceAPI):
    """Class which provide API to operate Dataset resource."""

    BACKEND_SERVICE_NAME = ServiceName.PAI_WORKSPACE

    _list_method = "list_datasets_with_options"
    _get_method = "get_dataset_with_options"
    _delete_method = "delete_dataset_with_options"
    _create_method = "create_dataset_with_options"

    def list(
        self,
        data_source_type: str = None,
        name: str = None,
        page_size: int = 20,
        page_number: int = 1,
        order: str = "DESC",
        **kwargs,
    ) -> PaginatedResult:
        """Returns Dataset in paging.

        Args:
            data_source_type: Data source type of the dataset using, support OSS and NAS.
            name: Name of the Dataset.
            page_number: Page number.
            page_size: Page size.
            order: Return list order.

        Returns:
            int: A list of datasets match the conditions.
        """

        request = ListDatasetsRequest(
            data_source_types=data_source_type,
            name=name,
            page_size=page_size,
            page_number=page_number,
            order=order,
            **kwargs,
        )

        resp: ListDatasetsResponseBody = self._do_request(
            self._list_method, request=request
        )

        return self.make_paginated_result(resp)

    def get_api_object_by_resource_id(
        self, resource_id: str
    ) -> Dict[str, Union[List[Dict[str, str]], str]]:
        result = self._do_request(self._get_method, dataset_id=resource_id)
        return result.to_map()

    def get(self, id: str) -> Dict[str, Any]:
        """Get Dataset by Id.

        Returns:
            dict:
        """

        return self.get_api_object_by_resource_id(resource_id=id)

    def delete(self, id):
        """Delete the Dataset."""
        self._do_request(self._delete_method, dataset_id=id)

    def create(
        self,
        uri: str,
        name: Optional[str] = None,
        data_source_type: str = None,
        options: Union[str, Dict[str, Any]] = None,
        description: str = None,
        labels: Optional[Dict[str, str]] = None,
        mount_path: str = "/mnt/data/",
        data_type: str = DataType.COMMON,
        accessibility: str = ResourceAccessibility.PUBLIC,
        property: str = FileProperty.DIRECTORY,
        source_type: str = DatasetSourceType.USER,
    ) -> str:
        """Create a Dataset resource."""

        labels = labels or dict()
        request = CreateDatasetRequest(
            accessibility=accessibility,
            data_source_type=data_source_type,
            data_type=data_type,
            description=description,
            labels=[
                DatasetLabel(key=key, value=value) for key, value in labels.items()
            ],
            name=name,
            options=self._patch_mount_path(options, mount_path=mount_path),
            property=property,
            source_type=source_type,
            uri=uri,
        )
        result = self._do_request(self._create_method, request=request)
        return result.dataset_id

    @classmethod
    def _patch_mount_path(cls, options: Union[str, Dict[str, Any]], mount_path: str):
        if isinstance(options, str):
            options = json.loads(options)
        options = options or dict()
        if mount_path:
            options.update({"mountPath": mount_path})

        return json.dumps(options)



#  Copyright 2023 Alibaba, Inc. or its affiliates.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from pai.pipeline.types.artifact import DataType


class ArchivedArtifact(object):
    """Archived artifact instance, hold the metadata, value and source of the artifact."""

    def __init__(self, metadata, value, name, producer_id, id):
        """ArchivedArtifact class constructor.

        Args:
            metadata (:class:`~.pai.pipeline.artifact.ArtifactMetadata`): Artifact metadata info.
            value (:class:`~.pai.pipeline.artifact.ArtifactValue`): Artifact value.
            name (str): name of artifact, actually, it is the name of artifact in manifest of
                source pipeline.
            producer_id (str): Producer of the artifact, identified by step name of pipeline.
            artifact_id (str): Unique artifact identifier in pipeline service.
        """
        self.name = name
        self.metadata = metadata
        self.value = value
        self.producer_id = producer_id
        self.id = id

    def __repr__(self):
        return "%s:Id=%s,Name=%s" % (
            type(self).__name__,
            self.id,
            self.name,
        )

    @classmethod
    def deserialize(cls, output):
        """Build new ArtifactInfo instance from Run output json.

        Output Example:

        {
          'Info': {
            'value': '{"location": {"table": "pai_temp_77c08aeb2e514c9d8649feba4a88ee77"}}',
            'metadata': {
              'path': '/tmp/outputs/artifacts/outputArtifact/data',
              'type': {
                'DataSet': {
                  'locationType': 'MaxComputeTable'
                }
              }
            }
          },
          'Name': 'outputArtifact',
          'Producer': 'flow-ec67rsug8kyly4049z',
          'CreateTime': 1595214018000,
          'Type': 'DataSet',
          'Id': 'artifact-e0xdqhsfhqctxkpyli'
        }

        Args:
            output (dict): Run output return by pipeline service.
        Returns:
            ArchivedArtifact: ArtifactInfo instance.
        """
        name = output["Name"]
        metadata = output["Info"]["metadata"]
        value = output["Info"]["value"]
        producer_id = output["Producer"]
        artifact_id = output["Id"]
        return ArchivedArtifact(
            metadata=metadata,
            value=value,
            name=name,
            producer_id=producer_id,
            id=artifact_id,
        )

    @property
    def is_model(self):
        return self.metadata.data_type == DataType.Model



#  Copyright 2023 Alibaba, Inc. or its affiliates.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import absolute_import

import json
import re
from abc import ABCMeta, abstractmethod

import six
from odps.df import DataFrame as ODPSDataFrame
from odps.models import Partition as ODPSPartition
from odps.models import Table as ODPSTable
from odps.models import Volume as ODPSVolume
from odps.models.ml.offlinemodel import OfflineModel as ODPSOfflineModel
from six import with_metaclass

from pai.common.oss_utils import OssUriObj
from pai.common.utils import is_iterable
from pai.pipeline.types.variable import PipelineVariable


class ArtifactMetadataUtils(object):
    """Util class using for create artifact metadata."""

    @staticmethod
    def maxc_table():
        """Return metadata represent a MaxComputeTable."""
        return LocationArtifactMetadata(
            data_type=DataType.DataSet,
            location_type=LocationType.MaxComputeTable,
        )

    @staticmethod
    def oss_dataset():
        return LocationArtifactMetadata(
            data_type=DataType.DataSet, location_type=LocationType.OSS
        )

    @staticmethod
    def maxc_offlinemodel():
        return LocationArtifactMetadata(
            data_type=DataType.Model, location_type=LocationType.MaxComputeOfflineModel
        )

    @staticmethod
    def maxc_volume():
        return LocationArtifactMetadata(
            data_type=DataType.DataSet, location_type=LocationType.MaxComputeVolume
        )

    @staticmethod
    def raw():
        return LocationArtifactMetadata(data_type=DataType.Any)


class DataType(object):
    DataSet = "DataSet"
    Model = "Model"
    Any = "Any"
    ModelEvaluation = "ModelEvaluation"


class LocationType(object):
    MaxComputeTable = "MaxComputeTable"
    MaxComputeVolume = "MaxComputeVolume"
    MaxComputeOfflineModel = "MaxComputeOfflineModel"
    OSS = "OSS"


class ModelType(object):
    OfflineModel = "OfflineModel"
    PMML = "PMML"


class PipelineArtifact(PipelineVariable):
    """Input/output artifact definition of the Pipeline and operator.

    Examples:

        from pai.operator.types import MetadataBuilder, PipelineArtifact
        from pai.operator import ContainerOperator

        op = ContainerOperator(
            image_uri="python:3",
            inputs=[
                PipelineArtifact(name="foo", metadata=MetadataBuilder.to_dict()),
            ]

        )
        pass

    """

    variable_category = "artifacts"

    def __init__(
        self,
        name,
        metadata=None,
        path=None,
        desc=None,
        io_type="inputs",
        value=None,
        from_=None,
        required=False,
        repeated=False,
        parent=None,
    ):
        super(PipelineArtifact, self).__init__(
            name=name,
            desc=desc,
            io_type=io_type,
            value=value,
            from_=from_,
            required=required,
            parent=parent,
        )
        self.metadata = metadata
        self.path = path
        self.repeated = repeated
        self._count = None

    @property
    def count(self):
        return self._count

    @count.setter
    def count(self, value):
        if not self.repeated:
            raise ValueError("no repeated artifact do not has count attribute.")

        if not isinstance(value, six.integer_types) or value <= 0:
            raise ValueError("invalid count for repeated artifact:%s" % value)

        self._count = value

    def reset_count(self):
        self._count = None

    def validate_from(self, arg):
        if isinstance(arg, PipelineArtifact):
            source = arg
            if source.repeated and not source.parent:
                raise ValueError("repeated artifact could not assign entirely.")
        elif isinstance(arg, PipelineArtifactElement):
            source = arg.artifact
        else:
            raise ValueError(
                "arg is expected to be type of 'PipelineParameter' "
                "but was actually of type '%s'" % type(arg)
            )

        if (
            source.metadata is not None
            and self.metadata is not None
            and source.metadata != self.metadata
        ):
            return False
        return True

    def assign(self, arg):
        def _validate(item):
            if isinstance(
                item, (PipelineArtifact, PipelineArtifactElement)
            ) and not self.validate_from(item):
                raise ValueError(
                    "invalid assignment. %s left: %s, right: %s"
                    % (self.fullname, self, arg)
                )
            elif not isinstance(
                item, (PipelineArtifact, PipelineArtifactElement)
            ) and not self.validate_value(item):
                raise ValueError("Value(%s) is invalid value for %s" % (item, self))

        if self.repeated:
            if is_iterable(arg):
                value = [item for item in arg]
            else:
                value = [arg]
            for item in value:
                _validate(item)

            self.value = value
        else:
            _validate(arg)
            if isinstance(arg, (PipelineArtifact, PipelineArtifactElement)):
                self.from_ = arg
            else:
                # We need to transform artifact input value input to standard artifact value
                # which PAIFlow service recognized.
                self.value = self._translate_value(arg)["value"]

    def __getitem__(self, key):
        if not self.repeated:
            raise KeyError("no repeated artifact is not indexable.")

        if isinstance(key, six.integer_types):
            return PipelineArtifactElement(self, key)
        elif isinstance(key, slice):
            if key.stop is None:
                raise ValueError(
                    "slice of repeated artifact should have stop property."
                )
            start, stop, step = key.indices(key.stop)
            indexes = range(stop)[start:stop:step]
            return [
                PipelineArtifactElement(artifact=self, index=idx) for idx in indexes
            ]

        raise KeyError("please provide integer to index the artifact element.")

    # TODO: Artifact value validation
    def validate_value(self, val):
        return True

    def normalized_name(self, index):
        return "%s_%s" % (self.name, index)

    def depend_steps(self):
        def _depend_step(value):
            if isinstance(value, PipelineArtifact) and value.parent:
                return value.parent
            elif isinstance(value, PipelineArtifactElement) and value.artifact.parent:
                return value.artifact.parent

        if self.from_:
            return filter(None, [_depend_step(self.from_)])
        elif self.repeated and self.value:
            return filter(None, [_depend_step(item) for item in self.value])

    def translate_argument(self, arg):
        argument = {"name": self.name}
        if self.repeated:
            arg_list = arg if isinstance(arg, list) else [arg]
            results = [
                {"name": self.normalized_name(idx), "value": self._translate_value(v)}
                for idx, v in enumerate(arg_list)
            ]
            argument["value"] = results
        else:
            art_value = self._translate_value(arg)
            argument["value"] = art_value["value"]
            argument["metadata"] = art_value["metadata"]

        return argument

    def _translate_value(self, val):
        if isinstance(self.metadata, ArtifactMetadataBase):
            md_val = self.metadata.to_dict()
        else:
            md_val = self.metadata

        try:
            af_value = LocationArtifactValue.from_resource(val)
            value = json.dumps(af_value.to_dict(), sort_keys=True)
        except ValueError:
            value = val

        if (
            isinstance(self.metadata, LocationArtifactMetadata)
            and self.metadata.is_raw()
        ):
            metadata = LocationArtifactValue.metadata_from_value(val)
            if metadata:
                md_val = metadata.to_dict()

        return {
            "value": value,
            "metadata": md_val,
        }

    def to_argument(self):
        argument = {"name": self.name}
        if self.repeated:
            values = []
            for idx, item in enumerate(self.value):
                if isinstance(item, (PipelineArtifact, PipelineArtifactElement)):
                    values.append(
                        {
                            "name": self.normalized_name(idx),
                            "from": item.enclosed_fullname,
                        }
                    )
                else:
                    values.append(
                        {
                            "name": self.normalized_name(idx),
                            "value": None,
                        }
                    )
            argument["value"] = values
        elif self.from_:
            argument["from"] = self.from_.enclosed_fullname
        else:
            argument["value"] = self.value
        return argument

    def to_dict(self):
        d = super(PipelineArtifact, self).to_dict()
        d["metadata"] = self.metadata.to_dict()
        if self.value is not None:
            if isinstance(self.value, LocationArtifactValue):
                d["value"] = self.value.to_dict()
            else:
                d["value"] = self.value
        if self.path is not None:
            d["path"] = self.path
        d["required"] = self.required
        if self.repeated:
            d["repeated"] = self.repeated
        return d


class PipelineArtifactElement(object):
    def __init__(self, artifact, index):
        self.artifact = artifact
        self.index = index
        self._from = None
        self._value = None

    @property
    def name(self):
        return "%s_%s" % (self.artifact.name, self.index)

    @property
    def fullname(self):
        return ".".join(
            [
                self.artifact.parent.ref_name,
                self.artifact.io_type,
                self.artifact.variable_category,
                self.name,
            ]
        )

    @property
    def enclosed_fullname(self):
        return "{{%s}}" % self.fullname

    @property
    def parent(self):
        return self.artifact.parent


class ArtifactMetadataBase(with_metaclass(ABCMeta)):
    @abstractmethod
    def to_dict(self):
        pass


class LocationArtifactMetadata(ArtifactMetadataBase):
    def __init__(self, data_type, location_type=None, type_attributes=None):
        self.data_type = data_type
        self.location_type = location_type
        self.type_attributes = type_attributes or dict()

    def __str__(self):
        return "%s:data_type=%s:location_type=%s" % (
            type(self).__name__,
            self.data_type,
            self.location_type,
        )

    def is_raw(self):
        return self.data_type == DataType.Any and self.location_type is None

    def to_dict(self):
        d = {
            "type": {
                self.data_type: {
                    "locationType": self.location_type,
                }
            }
        }

        d["type"][self.data_type].update(self.type_attributes)

        return d

    @property
    def value(self):
        return self.to_dict()

    def __eq__(self, other):
        if self.data_type == DataType.Any:
            return True
        elif isinstance(other, LocationArtifactMetadata):
            return other.data_type == DataType.Any or (
                self.data_type == other.data_type
                and self.location_type == other.location_type
                and self.type_attributes == other.type_attributes
            )
        elif isinstance(other, dict):
            other = self.from_dict(other)
            return self.__eq__(other)
        return False

    def __ne__(self, other):
        return not self.__eq__(other)

    @classmethod
    def from_dict(cls, d):
        af_typ = d["type"]
        data_type = list(af_typ.keys())[0]
        type_attributes = af_typ[data_type].copy()
        location_type = type_attributes.pop("locationType", None)
        return cls(
            data_type=data_type,
            location_type=location_type,
            type_attributes=type_attributes,
        )


class LocationArtifactValue(object):
    def __init__(self):
        pass

    @classmethod
    def from_resource(cls, resource, metadata=None):
        """Get artifact value from input resource and artifact metadata.

        Args:
            resource: Input resource which can be ODPS url, OSS url, PyODPS table object, etc.
            metadata: Metadata for the input artifact.

        Returns:
            LocationArtifactValue: A artifact value parsed from input resource.

        """
        from pai.pipeline.artifact import ArchivedArtifact

        if isinstance(resource, six.string_types):
            if resource.startswith("odps://"):
                val, _ = MaxComputeResourceArtifact.from_resource_url(resource)
                return val
            elif resource.startswith("oss://"):
                val, _ = OSSArtifact.from_resource_url(resource)
                return val
            else:
                try:
                    resource_dict = json.loads(resource)
                except json.JSONDecodeError:
                    raise ValueError("Not support artifact url schema:%s", resource)
                return cls.from_raw_value(resource_dict, metadata)
        elif isinstance(
            resource, (ODPSTable, ODPSPartition, ODPSDataFrame, ODPSVolume)
        ):
            return MaxComputeResourceArtifact.from_odps_resource(resource)
        elif isinstance(resource, ArchivedArtifact):
            return resource.value
        elif isinstance(resource, LocationArtifactValue):
            return resource

        raise ValueError("Not supported artifact resource:%s", type(resource))

    @classmethod
    def metadata_from_value(cls, resource):
        """Try to get metadata from raw input for the artifact.

        Args:
            resource: Input for the artifact.

        Returns:
            (ArtifactMetadataBase): Metadata for the input artifact value.

        """
        from pai.pipeline.artifact import ArchivedArtifact

        if isinstance(resource, six.string_types):
            if resource.startswith("odps://"):
                _, metadata = MaxComputeResourceArtifact.from_resource_url(resource)
            elif resource.startswith("oss://"):
                _, metadata = OSSArtifact.from_resource_url(resource)
            else:
                return None
        elif isinstance(resource, (ODPSTable, ODPSPartition, ODPSDataFrame)):
            metadata = ArtifactMetadataUtils.maxc_table()
        elif isinstance(resource, ArchivedArtifact):
            try:
                metadata = LocationArtifactMetadata.from_dict(resource.metadata)
            except (KeyError, ValueError):
                return None
        else:
            return None
        return metadata

    @classmethod
    def from_raw_value(cls, value, metadata):
        if metadata is None:
            raise ValueError(
                400,
                "ArtifactMetadata should provide while parse artifact value from dict data.",
            )
        if metadata.location_type == LocationType.OSS:
            return OSSArtifact.from_dict(value)
        elif metadata.location_type == LocationType.MaxComputeTable:
            return MaxComputeTableArtifact.from_dict(value)
        elif metadata.location_type == LocationType.MaxComputeOfflineModel:
            return MaxComputeOfflineModelArtifact.from_dict(value)
        elif metadata.location_type == LocationType.MaxComputeVolume:
            return MaxComputeVolumeArtifact.from_dict(value)
        else:
            raise ValueError(
                "Not support artifact location_type: %s", metadata.location_type
            )

    @classmethod
    def get_param_ref(cls, param):
        from pai.pipeline.types import PipelineParameter

        if not param:
            return

        elif isinstance(param, PipelineParameter):
            ref = param.enclosed_fullname
        elif isinstance(param, six.string_types):
            ref = "{{inputs.parameters.%s}}" % param
        else:
            raise ValueError(
                "expect PipelineParameter or string, but given %s", type(param)
            )
        return ref

    def to_dict(self):
        pass


class MaxComputeResourceArtifact(LocationArtifactValue):
    MaxComputeResourceUrlPattern = re.compile(
        r"odps://(?P<project>[^/]+)/(?P<resource_type>(?:tables)|(?:volumes)|(?:offlinemodels))/"
        r"(?P<resource_name>[^/]+)/?(?P<sub_resource>[^\?]+)?(?P<arguments>[.*])?"
    )

    def __init__(self, project, endpoint=None):
        super(MaxComputeResourceArtifact, self).__init__()
        self.project = project
        self.endpoint = endpoint

    def to_dict(self):
        d = {
            "location": {
                "project": self.project,
            }
        }
        if self.endpoint:
            d["location"]["endpoint"] = self.endpoint
        return d

    @classmethod
    def from_resource_url(cls, resource_url):
        """Parse MaxCompute(ODPS) resource in url schema and returns artifact value and metadata.

        Args:
            resource_url: An ODPS(MaxCompute) table, tablePartition, offline-model or volume in url schema.

        Returns:
            tuple: A tuple of  MaxCompute artifact value and artifact metadata.

        """
        matches = cls.MaxComputeResourceUrlPattern.match(resource_url)
        if not matches:
            raise ValueError("Not support MaxCompute resource url format.")
        resource_type = matches.group("resource_type")
        project = matches.group("project")
        if resource_type == "tables":
            table = matches.group("resource_name")
            partition = matches.group("sub_resource")
            return (
                MaxComputeTableArtifact(
                    project=project,
                    table=table,
                    partition=partition.strip("/") if partition else None,
                ),
                ArtifactMetadataUtils.maxc_table(),
            )
        elif resource_type == "volumes":
            volume = matches.group("resource_name")
            sub_resource = matches.group("sub_resource").strip("/")
            idx = sub_resource.find("/")
            partition = sub_resource[:idx]
            file_name = sub_resource[idx + 1 :]
            return (
                MaxComputeVolumeArtifact(
                    project=project,
                    volume=volume,
                    partition=partition,
                    file_name=file_name,
                ),
                ArtifactMetadataUtils.maxc_volume(),
            )

        elif resource_type == "offlinemodels":
            name = matches.group("resource_name")
            return (
                MaxComputeOfflineModelArtifact(
                    project=project,
                    offline_model=name,
                ),
                ArtifactMetadataUtils.maxc_offlinemodel(),
            )
        else:
            raise ValueError("Not support MaxCompute resource type :%s" % resource_type)

    @classmethod
    def from_odps_resource(cls, resource):
        if not isinstance(
            resource, (ODPSTable, ODPSPartition, ODPSDataFrame, ODPSVolume)
        ):
            raise ValueError("Not support resource type:%s" % type(resource))

        if isinstance(resource, ODPSDataFrame):
            resource = resource.data

        if isinstance(resource, ODPSTable):
            project = resource.project.name
            table = resource.name
            return MaxComputeTableArtifact(
                project=project,
                table=table,
            )
        elif isinstance(resource, ODPSPartition):
            table = resource.table.name
            project = resource.table.project.name
            partition = ",".join(
                ["%s=%s" % (k, v) for k, v in resource.partition_spec.kv.items()]
            )
            return MaxComputeTableArtifact(
                project=project, partition=partition, table=table
            )
        elif isinstance(resource, ODPSOfflineModel):
            project = resource.project.name
            offlinemodel = resource.name
            return MaxComputeOfflineModelArtifact(
                project=project, offline_model=offlinemodel
            )
        # TODO: ODPSVolume Support
        elif isinstance(resource, ODPSVolume):
            pass


class MaxComputeTableArtifact(MaxComputeResourceArtifact):
    def __init__(self, table, project, endpoint=None, partition=None):
        super(MaxComputeTableArtifact, self).__init__(
            project=project, endpoint=endpoint
        )
        self.table = table
        self.partition = partition

    def to_dict(self):
        d = super(MaxComputeTableArtifact, self).to_dict()
        d["location"]["table"] = self.table
        if self.partition:
            d["location"]["partition"] = self.partition
        return d

    @classmethod
    def from_dict(cls, d):
        table = d["location"]["table"]
        project = d["location"].get("project")
        endpoint = d["location"].get("endpoint")
        partition = d["location"].get("partition")
        return cls(table=table, project=project, endpoint=endpoint, partition=partition)

    @classmethod
    def value_from_param(cls, table_name, partition=None):
        table_ref = cls.get_param_ref(table_name)
        partition_ref = cls.get_param_ref(partition)
        if not table_ref:
            raise ValueError("MaxComputeTableArtifact require table not be None")
        d = {
            "location": {
                "table": table_ref,
            }
        }

        if partition_ref is not None:
            d["location"]["partition"] = partition_ref
        return json.dumps(d)


class MaxComputeOfflineModelArtifact(MaxComputeResourceArtifact):
    def __init__(self, offline_model, project, endpoint=None):
        super(MaxComputeOfflineModelArtifact, self).__init__(
            project=project, endpoint=endpoint
        )
        self.offline_model = offline_model

    def to_dict(self):
        d = super(MaxComputeOfflineModelArtifact, self).to_dict()
        d["location"]["name"] = self.offline_model
        d["name"] = self.offline_model
        return d

    @classmethod
    def from_dict(cls, d):
        project = d["location"]["project"]
        name = d["location"]["name"]
        endpoint = d["location"].get("endpoint")
        return cls(offline_model=name, project=project, endpoint=endpoint)


class MaxComputeVolumeArtifact(MaxComputeResourceArtifact):
    def __init__(self, volume, project, file_name, endpoint=None, partition=None):
        super(MaxComputeVolumeArtifact, self).__init__(
            project=project, endpoint=endpoint
        )
        self.volume = volume
        self.partition = partition
        self.file_name = file_name

    def to_dict(self):
        d = super(MaxComputeVolumeArtifact, self).to_dict()
        d["location"]["volume"] = self.volume
        d["location"]["volumePartition"] = self.partition
        d["location"]["file"] = self.file_name
        return d

    @classmethod
    def from_dict(cls, d):
        endpoint = d["location"].get("endpoint", None)
        project = d["location"]["project"]
        volume = d["location"]["volume"]
        partition = d["location"].get("volumePartition", None)
        file = d["location"].get("file", None)
        return cls(
            volume=volume,
            project=project,
            file_name=file,
            endpoint=endpoint,
            partition=partition,
        )


class OSSArtifact(LocationArtifactValue):
    """OSSArtifact instance represent a OSS artifact value."""

    def __init__(self, bucket, key, endpoint, role_arn=None):
        """

        Args:
            bucket: OSS bucket name.
            key: OSS object key.
            endpoint: Endpoint for the OSS bucket.
            role_arn: RoleArn used for OSS access.
        """
        super(OSSArtifact, self).__init__()
        self.bucket = bucket
        self.endpoint = endpoint
        self.role_arn = role_arn
        self.key = key

    def to_dict(self):
        d = {
            "location": {
                "endpoint": self.endpoint,
                "bucket": self.bucket,
                "key": self.key,
            }
        }

        if self.role_arn:
            d["location"]["rolearn"] = self.role_arn

        return d

    @classmethod
    def from_resource_url(cls, resource):
        """Initialize a OSSArtifact instance from URL in OSS schema.
        Args:
            resource: URL in OSS schema.
        Returns:
            OSSArtifact instance.
        """
        bucket_name, object_key, endpoint, role_arn = OssUriObj.parse(resource)
        return (
            cls(
                bucket=bucket_name, key=object_key, endpoint=endpoint, role_arn=role_arn
            ),
            ArtifactMetadataUtils.oss_dataset(),
        )

    @classmethod
    def from_dict(cls, d):
        if not d["location"].get("bucket"):
            return
        bucket = d["location"]["bucket"]
        key = d["location"]["key"]
        endpoint = d["location"].get("endpoint")
        rolearn = d["location"].get("rolearn")
        return OSSArtifact(bucket=bucket, key=key, endpoint=endpoint, role_arn=rolearn)



# coding: utf-8

#  Copyright 2023 Alibaba, Inc. or its affiliates.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import print_function

import uuid

import six
from deprecated import deprecated

from ...common.logging import get_logger
from ...common.yaml_utils import dump as yaml_dump
from ...session import get_default_session
from ..types.variable import PipelineVariable
from ._base import UnRegisteredComponent

PAI_MANIFEST_SPEC_INPUTS_ENV_KEY = "PAI_MANIFEST_SPEC_INPUTS"
PAI_MANIFEST_SPEC_OUTPUTS_ENV_KEY = "PAI_MANIFEST_SPEC_OUTPUTS"
PAI_INPUTS_PARAMETERS_ENV_KEY = "PAI_INPUTS_PARAMETERS"

_logger = get_logger(__name__)


@deprecated(
    reason="ContainerComponent is deprecated and will be removed in the future."
)
class ContainerComponent(UnRegisteredComponent):
    def __init__(
        self,
        image_uri,
        command,
        args=None,
        image_registry_config=None,
        inputs=None,
        outputs=None,
        env=None,
    ):
        self._image_uri = image_uri
        self._image_registry_config = image_registry_config
        self._command = command
        self._args = args
        self._env = env
        self._guid = uuid.uuid4().hex

        super(ContainerComponent, self).__init__(
            inputs=inputs,
            outputs=outputs,
        )

    @classmethod
    def _transform_env(cls, env):
        if not env:
            return dict()
        return {
            k: v.enclosed_fullname if isinstance(v, PipelineVariable) else str(v)
            for k, v in env.items()
        }

    @classmethod
    def _transform_commands(cls, commands):
        if isinstance(commands, six.string_types):
            return [commands]
        if not commands:
            return []

        return [
            c.enclosed_fullname if isinstance(c, PipelineVariable) else c
            for c in commands
        ]

    def to_dict(self, identifier=None, version=None):
        d = super(ContainerComponent, self).to_dict()

        if identifier is not None:
            d["metadata"]["identifier"] = identifier
        if version is not None:
            d["metadata"]["version"] = version

        if get_default_session():
            d["metadata"]["provider"] = get_default_session().provider

        d["spec"]["container"] = {
            "image": self._image_uri,
            "command": self._transform_commands(self._command),
        }
        d["spec"]["container"]["imageRegistryConfig"] = (
            self._image_registry_config or dict()
        )
        if self._env:
            d["spec"]["container"]["envs"] = self._transform_env(self._env or dict())

        if self._args:
            d["spec"]["container"]["args"] = self._transform_commands(self._args)
        return d

    def to_manifest(self, identifier, version):
        return yaml_dump(self.to_dict(identifier=identifier, version=version))



#  Copyright 2023 Alibaba, Inc. or its affiliates.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import json
import typing
from typing import Any, Dict, Union

from ..common.logging import get_logger
from ..libs.alibabacloud_eas20210701.models import (
    CreateServiceRequest,
    CreateServiceResponseBody,
    DescribeMachineSpecRequest,
    DescribeMachineSpecResponseBody,
    ListServicesRequest,
    ListServicesResponseBody,
    ReleaseServiceRequest,
    UpdateServiceRequest,
    UpdateServiceVersionRequest,
)
from .base import PaginatedResult, ResourceAPI, ServiceName

logger = get_logger(__name__)


class ServiceAPI(ResourceAPI):
    BACKEND_SERVICE_NAME = ServiceName.PAI_EAS

    _create_method = "create_service_with_options"
    _update_method = "update_service_with_options"
    _update_version_method = "update_service_version_with_options"
    _list_method = "list_services_with_options"
    _get_method = "describe_service_with_options"
    _start_method = "start_service_with_options"
    _stop_method = "stop_service_with_options"
    _delete_method = "delete_service_with_options"
    _release_method = "release_service_with_options"

    _get_group_method = "describe_group_with_options"
    _list_groups_method = "list_group_with_options"
    _describe_machine_method = "describe_machine_spec_with_options"

    def __init__(self, region_id, acs_client, **kwargs):
        super(ServiceAPI, self).__init__(acs_client=acs_client, **kwargs)
        self.region_id = region_id

    def list(
        self, filter=None, order=None, page_number=None, page_size=None, sort=None
    ) -> PaginatedResult:
        """

        Returns:

        """
        request = ListServicesRequest(
            filter=filter,
            order=order,
            page_number=page_number,
            page_size=page_size,
            sort=sort,
        )
        resp: ListServicesResponseBody = self._do_request(self._list_method, request)
        return self.make_paginated_result(resp)

    def get_api_object_by_resource_id(self, resource_id):
        resp = self._do_request(
            self._get_method,
            cluster_id=self.region_id,
            service_name=resource_id,
        )
        return resp.to_map()

    def get(self, name: str) -> Dict[str, Any]:
        return self.get_api_object_by_resource_id(resource_id=name)

    def create(
        self, config: Union[str, typing.Dict], labels: Dict[str, str] = None
    ) -> str:
        """Create a PAI EAS Service resource, returns the service name."""

        if isinstance(config, str):
            config_obj = json.loads(config)
        else:
            config_obj = config

        request = CreateServiceRequest(body=config_obj, labels=labels)
        resp: CreateServiceResponseBody = self._do_request(self._create_method, request)
        return resp.service_name

    def release(self, name, weight):
        request = ReleaseServiceRequest(weight=weight)
        self._do_request(
            self._release_method,
            cluster_id=self.region_id,
            service_name=name,
            request=request,
        )

    def start(self, name):
        self._do_request(
            self._start_method, cluster_id=self.region_id, service_name=name
        )

    def stop(self, name):
        self._do_request(
            self._stop_method, cluster_id=self.region_id, service_name=name
        )

    def delete(self, name):
        self._do_request(
            self._delete_method, cluster_id=self.region_id, service_name=name
        )

    def update(self, name, config: Union[str, typing.Dict]):

        if isinstance(config, str):
            config_obj = json.loads(config)
        else:
            config_obj = config

        request: UpdateServiceRequest = UpdateServiceRequest(
            body=config_obj,
        )

        self._do_request(
            self._update_method,
            cluster_id=self.region_id,
            service_name=name,
            request=request,
        )

    def update_version(self, name, version):
        request = UpdateServiceVersionRequest(version=version)
        self._do_request(
            self._update_version_method,
            cluster_id=self.region_id,
            service_name=name,
            request=request,
        )

    def get_group(self, group_name) -> Dict[str, Any]:
        resp = self._do_request(
            self._get_group_method,
            cluster_id=self.region_id,
            group_name=group_name,
        )
        return resp.to_map()

    def describe_machine(self) -> Dict[str, Any]:
        """List the machine spec supported by PAI-EAS."""
        # DescribeMachineSpec API always returns all supported machine specs.
        request = DescribeMachineSpecRequest()
        resp: DescribeMachineSpecResponseBody = self._do_request(
            self._describe_machine_method,
            request,
        )
        return resp.to_map()
